import logging
from argparse import ArgumentParser

from castor_extractor.knowledge import confluence  # type: ignore
from castor_extractor.utils import parse_filled_arguments  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    parser.add_argument("-a", "--account_id", help="Confluence account id")
    parser.add_argument("-b", "--base_url", help="Confluence account base url")
    parser.add_argument("-o", "--output", help="Directory to write to")
    parser.add_argument("-t", "--token", help="Confluence API token")
    parser.add_argument("-u", "--username", help="Confluence username")

    parser.add_argument(
        "--include-archived-spaces",
        action="store_true",
        default=False,
        help="Include pages from archived spaces (Optional)",
    )
    parser.add_argument(
        "--include-personal-spaces",
        action="store_true",
        default=False,
        help="Include pages from personal spaces (Optional)",
    )
    parser.add_argument(
        "--space-ids-allowed",
        type=str,
        nargs="+",
        help=(
            "List of Confluence space IDs allowed for extraction (Optional). "
            "Only pages from these Spaces will be extracted. "
            "This overrides any other filtering (archived, personal, etc.)"
        ),
    )
    parser.add_argument(
        "--space-ids-blocked",
        type=str,
        nargs="+",
        help="List of Confluence space IDs to exclude fom the extraction (Optional)",
    )

    confluence.extract_all(**parse_filled_arguments(parser))



from enum import Enum


class LineageEntity(Enum):
    """Entities that can be linked in Databricks lineage"""

    COLUMN = "COLUMN"
    TABLE = "TABLE"


class TagEntity(Enum):
    """Entities that can be tagged in Databricks"""

    COLUMN = "COLUMN"
    TABLE = "TABLE"



from typing import Any, Union, cast

Field = Union[str, dict[str, Any]]
Fields = Union[Field, tuple[Field, ...]]


def format_fields(fields: Fields) -> str:
    """format field for looker api"""

    if isinstance(fields, str):
        return fields

    if isinstance(fields, dict):
        formatted = {
            k: format_fields(cast(Fields, v)) for k, v in fields.items()
        }
        return ", ".join(f"{k}({v})" for k, v in formatted.items())

    if isinstance(fields, tuple):
        return ", ".join(format_fields(f) for f in fields)

    raise ValueError(f"Expected tuple, dict or string, got: {fields}")



from datetime import date, datetime
from typing import Optional

from ....utils import at_midnight, clean_path, format_date, yesterday


def _time_filter(day: Optional[date]) -> tuple[datetime, datetime]:
    target_day = day or yesterday()
    start = at_midnight(target_day)
    end = datetime.combine(target_day, datetime.max.time())
    return start, end


class PowerBiEndpointFactory:
    def __init__(self, login_url: str, api_base: str):
        self.app_base = clean_path(login_url)
        self.rest_api_base = clean_path(api_base)

    def activity_events(self, day: Optional[date]) -> str:
        start, end = _time_filter(day)
        url = f"{self.rest_api_base}/admin/activityevents"
        url += "?$filter=Activity eq 'viewreport'"
        url += f"&startDateTime='{format_date(start)}'"
        url += f"&endDateTime='{format_date(end)}'"
        return url

    def authority(self, tenant_id: str) -> str:
        return f"{self.app_base}/{tenant_id}"

    def dashboards(self) -> str:
        return f"{self.rest_api_base}/admin/dashboards"

    def datasets(self) -> str:
        return f"{self.rest_api_base}/admin/datasets"

    def groups(self) -> str:
        return f"{self.rest_api_base}/admin/groups"

    def metadata_create_scan(self) -> str:
        return f"{self.rest_api_base}/admin/workspaces/getInfo"

    def metadata_scan_result(self, scan_id: int) -> str:
        return f"{self.rest_api_base}/admin/workspaces/scanResult/{scan_id}"

    def metadata_scan_status(self, scan_id: int) -> str:
        return f"{self.rest_api_base}/admin/workspaces/scanStatus/{scan_id}"

    def pages(self, report_id: str) -> str:
        return f"{self.rest_api_base}/admin/reports/{report_id}/pages"

    def reports(self) -> str:
        return f"{self.rest_api_base}/admin/reports"

    def workspace_ids(self) -> str:
        return f"{self.rest_api_base}/admin/workspaces/modified"



import logging
from collections.abc import Iterator
from datetime import date
from functools import partial
from http import HTTPStatus
from time import sleep
from typing import Optional, Union

import requests
from requests import HTTPError

from ....utils import (
    APIClient,
    fetch_all_pages,
    retry_request,
)
from ..assets import PowerBiAsset
from .authentication import PowerBiBearerAuth
from .constants import Keys
from .credentials import PowerbiCredentials
from .endpoints import PowerBiEndpointFactory
from .pagination import PowerBiPagination

POWERBI_DEFAULT_TIMEOUT_S = 30
# The route we use to fetch workspaces info can retrieve a maximum of
# 100 workspaces per call
# More: https://learn.microsoft.com/en-us/rest/api/power-bi/admin/workspace-info-post-workspace-info#request-body
METADATA_BATCH_SIZE = 100
POWERBI_SCAN_STATUS_DONE = "Succeeded"
POWERBI_SCAN_SLEEP_S = 1
POWERBI_SCAN_TIMEOUT_S = 60

MAX_RETRY_PAGES = 1
RETRY_PAGES_TIMEOUT_MS = 35 * 1000  # 35 seconds

KEYS_TO_HIDE = ("ClientIP", "UserAgent")

logger = logging.getLogger(__name__)


class PowerbiClient(APIClient):
    def __init__(
        self,
        credentials: PowerbiCredentials,
    ):
        auth = PowerBiBearerAuth(credentials=credentials)
        super().__init__(
            auth=auth,
            timeout=POWERBI_DEFAULT_TIMEOUT_S,
        )
        self.endpoint_factory = PowerBiEndpointFactory(
            login_url=credentials.login_url,
            api_base=credentials.api_base,
        )

    def _activity_events(self, day: Optional[date] = None) -> Iterator[dict]:
        """
        Returns a list of activity events for the organization.
        https://learn.microsoft.com/en-us/power-bi/admin/service-admin-auditing#activityevents-rest-api
        - when no day is specified, fallback is yesterday
        """
        request = partial(
            self._get,
            endpoint=self.endpoint_factory.activity_events(day),
        )
        for event in fetch_all_pages(request, PowerBiPagination):
            for key in KEYS_TO_HIDE:
                if key in event:
                    del event[key]
            yield event

    def _datasets(self) -> Iterator[dict]:
        """
        Returns a list of datasets for the organization.
        https://learn.microsoft.com/en-us/rest/api/power-bi/admin/datasets-get-datasets-as-admin
        """
        yield from self._get(self.endpoint_factory.datasets())[Keys.VALUE]

    def _dashboards(self) -> Iterator[dict]:
        """
        Returns a list of dashboards for the organization.
        https://learn.microsoft.com/en-us/rest/api/power-bi/admin/dashboards-get-dashboards-as-admin
        """
        yield from self._get(self.endpoint_factory.dashboards())[Keys.VALUE]

    @retry_request(
        status_codes=(HTTPStatus.TOO_MANY_REQUESTS,),
        max_retries=MAX_RETRY_PAGES,
        base_ms=RETRY_PAGES_TIMEOUT_MS,
    )
    def _pages(self, report_id: str) -> Iterator[dict]:
        """
        Extracts the pages of a report.
        This endpoint is very flaky and frequently returns 400 and 404 errors.
        After around 50 requests, it hits the rate limit and returns 429 Too Many Requests,
        which is why we retry it after a short delay.
        Timeouts are also common; we must skip them because the extraction task
        might take too long otherwise.
        """
        pages_endpoint = self.endpoint_factory.pages(report_id)
        return self._get(pages_endpoint, retry_on_timeout=False)[Keys.VALUE]

    def _reports(self) -> Iterator[dict]:
        """
        Returns a list of reports for the organization.
        https://learn.microsoft.com/en-us/rest/api/power-bi/admin/reports-get-reports-as-admin
        """
        reports_endpoint = self.endpoint_factory.reports()
        reports = self._get(reports_endpoint)[Keys.VALUE]

        for report in reports:
            report_id = report.get(Keys.ID)

            try:
                pages = self._pages(report_id)
                report["pages"] = pages
            except (requests.HTTPError, requests.exceptions.Timeout) as e:
                logger.debug(e)
                continue

        return reports

    def _workspace_ids(self) -> list[str]:
        """
        Get workspaces ids from powerBI admin API.
        more: https://learn.microsoft.com/en-us/rest/api/power-bi/admin/workspace-info-get-modified-workspaces
        """
        params: dict[str, Union[bool, str]] = {
            Keys.INACTIVE_WORKSPACES: True,
            Keys.PERSONAL_WORKSPACES: True,
        }

        response = self._get(
            self.endpoint_factory.workspace_ids(),
            params=params,
        )

        return [x[Keys.ID] for x in response]

    def _get_scan_result(self, scan_id: int) -> Iterator[dict]:
        endpoint = self.endpoint_factory.metadata_scan_result(scan_id)
        yield from self._get(endpoint)[Keys.WORKSPACES]

    def _wait_for_scan_result(self, scan_id: int) -> bool:
        """
        Periodically checks the status of the metadata scan until the results
        are ready.
        """
        endpoint = self.endpoint_factory.metadata_scan_status(scan_id)
        total_waiting_time_s = 0

        while total_waiting_time_s < POWERBI_SCAN_TIMEOUT_S:
            try:
                result = self._get(endpoint)
            except HTTPError as e:
                logger.error(f"Scan {scan_id} failed. Error: {e}")
                return False

            if result[Keys.STATUS] == POWERBI_SCAN_STATUS_DONE:
                logger.info(f"scan {scan_id} ready")
                return True

            total_waiting_time_s += POWERBI_SCAN_SLEEP_S
            logger.info(
                f"Waiting {POWERBI_SCAN_SLEEP_S} sec for scan {scan_id} to be ready…",
            )
            sleep(POWERBI_SCAN_SLEEP_S)

        logger.warning(f"Scan {scan_id} timed out")
        return False

    def _create_scan(self, workspaces_ids: list[str]) -> int:
        """
        Tells the Power BI API to start an asynchronous metadata scan.
        Returns the scan's ID.
        https://learn.microsoft.com/en-us/rest/api/power-bi/admin/workspace-info-post-workspace-info
        """
        params = {
            "datasetExpressions": True,
            "datasetSchema": True,
            "datasourceDetails": True,
            "getArtifactUsers": True,
            "lineage": True,
        }
        request_body = {"workspaces": workspaces_ids}
        scan_id = self._post(
            self.endpoint_factory.metadata_create_scan(),
            params=params,
            data=request_body,
        )
        return scan_id[Keys.ID]

    def _metadata(self) -> Iterator[dict]:
        """
        Fetch metadata by workspace. The metadata scanning is asynchronous and
        requires the following steps:
        - create the asynchronous scan
        - periodically check the scan status to know when it's finished
        - get the actual scan results
        https://learn.microsoft.com/en-us/power-bi/enterprise/service-admin-metadata-scanning
        """
        ids = self._workspace_ids()

        for index in range(0, len(ids), METADATA_BATCH_SIZE):
            batch_ids = ids[index : index + METADATA_BATCH_SIZE]
            scan_id = self._create_scan(batch_ids)
            self._wait_for_scan_result(scan_id)
            yield from self._get_scan_result(scan_id)

    def test_connection(self) -> None:
        """Use credentials & verify requesting the API doesn't raise an error"""
        self._auth.refresh_token()

    def fetch(
        self,
        asset: PowerBiAsset,
        *,
        day: Optional[date] = None,
    ) -> Iterator[dict]:
        """
        Given a PowerBi asset, returns the corresponding data using the
        appropriate client.
        """
        if asset == PowerBiAsset.ACTIVITY_EVENTS:
            yield from self._activity_events(day=day)

        elif asset == PowerBiAsset.DATASETS:
            yield from self._datasets()

        elif asset == PowerBiAsset.DASHBOARDS:
            yield from self._dashboards()

        elif asset == PowerBiAsset.REPORTS:
            yield from self._reports()

        elif asset == PowerBiAsset.METADATA:
            yield from self._metadata()

        else:
            raise ValueError(f"This asset {asset} is unknown")



import itertools
import logging
from typing import Optional

from google.api_core.exceptions import Forbidden  # type: ignore
from google.cloud.bigquery import Client as GoogleCloudClient  # type: ignore
from google.cloud.bigquery.dataset import Dataset  # type: ignore
from google.oauth2.service_account import Credentials  # type: ignore

from ...utils import SqlalchemyClient, retry
from .types import SetTwoString

logger = logging.getLogger(__name__)

BIGQUERY_URI = "bigquery://"

CREDENTIALS_INFO_KEY = "credentials_info"
PROJECT_ID_KEY = "project_id"

_RETRY_NUMBER = 1
_RETRY_BASE_MS = 60_000


class BigQueryClient(SqlalchemyClient):
    """Connect to BigQuery and run SQL queries"""

    def __init__(
        self,
        credentials: dict,
        db_allowed: Optional[set[str]] = None,
        db_blocked: Optional[set[str]] = None,
        dataset_blocked: Optional[set[str]] = None,
    ):
        super().__init__(credentials)
        self._db_allowed = db_allowed
        self._db_blocked = db_blocked
        self._dataset_blocked = dataset_blocked
        self.credentials = self._credentials()
        self.client = self._client()
        self._projects: list[str] | None = None
        self._datasets: list[Dataset] | None = None

    @staticmethod
    def name() -> str:
        return "BigQuery"

    def _keep_project(self, project: str) -> bool:
        if self._db_allowed and project not in self._db_allowed:
            return False
        if self._db_blocked and project in self._db_blocked:
            return False
        return True

    def _keep_dataset(self, dataset: str) -> bool:
        if not self._dataset_blocked:
            return True

        return dataset not in self._dataset_blocked

    def _engine_options(self, credentials: dict) -> dict:
        return {
            CREDENTIALS_INFO_KEY: credentials,
        }

    def _build_uri(self, credentials: dict) -> str:
        return BIGQUERY_URI

    def _credentials(self) -> Credentials:
        assert CREDENTIALS_INFO_KEY in self._options, (
            "Missing BigQuery credentials in engine's options"
        )
        credentials = self._options[CREDENTIALS_INFO_KEY]
        return Credentials.from_service_account_info(credentials)

    def _client(self) -> GoogleCloudClient:
        return GoogleCloudClient(
            project=self._options[CREDENTIALS_INFO_KEY].get(PROJECT_ID_KEY),
            credentials=self.credentials,
        )

    def _list_datasets(self) -> list[Dataset]:
        """
        Returns datasets available for the given GCP client
        Cache the result in self._datasets to reduce number of API calls
        """
        if self._datasets is None:
            self._datasets = [
                dataset
                for project_id in self.get_projects()
                for dataset in self.client.list_datasets(project_id)
                if self._keep_dataset(dataset.dataset_id)
            ]
        return self._datasets

    @retry(
        (Forbidden,),
        max_retries=_RETRY_NUMBER,
        base_ms=_RETRY_BASE_MS,
        log_exc_info=True,
    )
    def get_projects(self) -> list[str]:
        """
        Returns distinct project_id available for the given GCP client
        Cache the result in self._projects to reduce number of API calls.

        Note: Calling list_projects from GoogleCloudClient causes some
        ```
        google.api_core.exceptions.Forbidden: 403 GET https://bigquery.googleapis.com/bigquery/v2/projects?prettyPrint=false
        Quota exceeded: Your user exceeded quota for concurrent project.lists requests.
        ````
        In that case, task should retry automatically.
        """
        if self._projects is None:
            self._projects = [
                p.project_id
                for p in self.client.list_projects(retry=None)  # type: ignore
                if self._keep_project(p.project_id)
            ]
        return self._projects

    def get_regions(self) -> SetTwoString:
        """
        Returns (project_id, region) available for the given GCP client
        - Loops trough projects -> datasets -> region
        - Returns distinct values
        Example:
            project_A
                -> dataset_1:US
            project_B
                -> empty
            project_C
                -> dataset_2:EU
                -> dataset_3:EU
        Will return:
            { (p_A, US), (p_C, EU) }
        """
        return {
            (ds.project, ds._properties["location"])
            for ds in self._list_datasets()
        }

    def get_extended_regions(self) -> SetTwoString:
        """
        Returns all combinations of (project_id, region) for the given client
        - Fetch all projects
        - Fetch all regions (cross projects)
        - Returns a combination of the two lists
        Example:
            project_A
                -> dataset_1:US
            project_B
                -> empty
            project_C
                -> dataset_2:EU
                -> dataset_3:EU
        Will return:
            { (p_A, EU), (p_A, US), (p_B, EU), (p_B, US), (p_C, EU), (p_C, US) }
        """
        projects = self.get_projects()
        regions = {ds._properties["location"] for ds in self._list_datasets()}
        combinations = itertools.product(projects, regions)
        return set(combinations)

    def get_datasets(self) -> SetTwoString:
        """
        Returns distinct (project_id, dataset_id) available for the given GCP client
        """
        return {(ds.project, ds.dataset_id) for ds in self._list_datasets()}



from ...types import ExternalAsset, classproperty


class LookerStudioAsset(ExternalAsset):
    ASSETS = "assets"
    SOURCE_QUERIES = "source_queries"
    VIEW_ACTIVITY = "view_activity"

    @classproperty
    def optional(cls) -> set["LookerStudioAsset"]:
        return {LookerStudioAsset.VIEW_ACTIVITY}



class LookerStudioAPIEndpoint:
    BASE_PATH = "https://datastudio.googleapis.com"

    @classmethod
    def search(cls) -> str:
        """
        Search a user's assets.
        See https://developers.google.com/looker-studio/integrate/api/reference/assets/search
        """
        return f"{cls.BASE_PATH}/v1/assets:search"

    @classmethod
    def permissions(cls, asset_name: str) -> str:
        """
        Get the permissions of an asset. The user must be the owner of the asset.
        See https://developers.google.com/looker-studio/integrate/api/reference/permissions/get
        """
        return f"{cls.BASE_PATH}/v1/assets/{asset_name}/permissions"



from argparse import ArgumentParser

from castor_extractor.utils import parse_filled_arguments  # type: ignore
from castor_extractor.visualization import looker  # type: ignore


def main():
    parser = ArgumentParser()
    parser.add_argument("-b", "--base-url", help="Looker base url")
    parser.add_argument("-c", "--client-id", help="Looker client id")
    parser.add_argument("-s", "--client-secret", help="Looker client secret")
    parser.add_argument("-o", "--output", help="Directory to write to")
    parser.add_argument("-t", "--timeout", type=int, help="Timeout in seconds")
    parser.add_argument(
        "--thread-pool-size",
        type=int,
        help="Thread pool size, if searching per folder",
    )
    parser.add_argument(
        "--safe-mode",
        "-S",
        help="Looker safe mode",
        action="store_true",
    )
    parser.add_argument(
        "--log-to-stdout",
        help="Send all log outputs to stdout instead of stderr",
        action="store_true",
    )

    parser.add_argument(
        "--search-per-folder",
        help="Fetches Looks and Dashboards per folder",
        action="store_true",
    )

    looker.extract_all(**parse_filled_arguments(parser))



from ...types import ExternalAsset


class CoalesceAsset(ExternalAsset):
    """Coalesce assets"""

    NODES = "nodes"


class CoalesceQualityAsset(ExternalAsset):
    """
    Coalesce Quality Assets
    Remark: having a dedicated Enum for Quality simplifies the process of
    searching pushed files
    """

    NODES = "nodes"
    RUN_RESULTS = "run_results"



class TableauApiError(ValueError):
    def __init__(self, error: str):
        super().__init__(f"Tableau API returned the following error: {error}")


class TableauApiTimeout(ValueError):
    def __init__(self, error: str):
        super().__init__(f"Tableau API returned a timeout error: {error}")



import logging
from collections.abc import Iterator
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from http import HTTPStatus
from typing import Callable, Iterable, Optional

from pydantic import BaseModel

from ....utils import (
    APIClient,
    RequestSafeMode,
    fetch_all_pages,
    retry,
)
from ..assets import SigmaAsset
from .authentication import SigmaBearerAuth
from .credentials import SigmaCredentials
from .endpoints import SigmaEndpointFactory
from .pagination import (
    SIGMA_API_LIMIT,
    SIGMA_QUERIES_PAGINATION_LIMIT,
    SigmaPagination,
)
from .sources_transformer import SigmaSourcesTransformer

logger = logging.getLogger(__name__)

_CONTENT_TYPE = "application/x-www-form-urlencoded"

_DATA_ELEMENTS: tuple[str, ...] = (
    "input-table",
    "pivot-table",
    "table",
    "visualization",
    "viz",
)

_SIGMA_TIMEOUT_S = 300

_SIGMA_HEADERS = {
    "Content-Type": _CONTENT_TYPE,
}

_VOLUME_IGNORED = 10_000
_IGNORED_ERROR_CODES = (
    HTTPStatus.BAD_REQUEST,
    HTTPStatus.BAD_GATEWAY,
    HTTPStatus.INTERNAL_SERVER_ERROR,
    HTTPStatus.CONFLICT,
    HTTPStatus.NOT_FOUND,
    HTTPStatus.FORBIDDEN,
)
SIGMA_SAFE_MODE = RequestSafeMode(
    max_errors=_VOLUME_IGNORED,
    status_codes=_IGNORED_ERROR_CODES,
)
SIGMA_SAFE_MODE_LINEAGE = RequestSafeMode(
    max_errors=_VOLUME_IGNORED,
    status_codes=(
        *_IGNORED_ERROR_CODES,
        HTTPStatus.FORBIDDEN,
    ),
)
_THREADS_LINEAGE = 10  # empirically found; hit the rate limit with 20 workers
_RETRY_NUMBER = 1
_RETRY_BASE_MS = 60_000


class LineageContext(BaseModel):
    """all info needed to build the endpoint for lineage retrieval"""

    workbook_id: str
    element_id: str


class Lineage(BaseModel):
    """holds response from lineage API and context used to retrieve it"""

    lineage: dict
    context: LineageContext


class SigmaClient(APIClient):
    def __init__(
        self,
        credentials: SigmaCredentials,
        safe_mode: Optional[RequestSafeMode] = None,
    ):
        auth = SigmaBearerAuth(
            host=credentials.host,
            token_payload=credentials.token_payload,
        )
        super().__init__(
            host=credentials.host,
            auth=auth,
            headers=_SIGMA_HEADERS,
            timeout=_SIGMA_TIMEOUT_S,
            safe_mode=safe_mode or SIGMA_SAFE_MODE,
        )

    def _get_paginated(
        self,
        endpoint: str,
        limit: int = SIGMA_API_LIMIT,
    ) -> Callable:
        """
        Sigma’s API does not experience random timeouts, unlike some other APIs.
        However, extracting queries from certain workbooks can take a
        significant amount of time.
        Previously, when a timeout occurred, the system would retry multiple
        times — even though we knew it would eventually fail due to the inherent
        slowness of the operation.
        These retries only delayed the inevitable failure without adding value.
        To address this, we've disabled retries on timeout and instead adjusted
        the page size when extracting queries.
        """
        return partial(
            self._get,
            retry_on_timeout=False,  # explained in the docstring
            endpoint=endpoint,
            params={"limit": limit},
        )

    def _get_all_datamodels(self) -> Iterator[dict]:
        request = self._get_paginated(
            endpoint=SigmaEndpointFactory.datamodels()
        )
        yield from fetch_all_pages(request, SigmaPagination)

    def _get_all_datasets(self) -> Iterator[dict]:
        request = self._get_paginated(endpoint=SigmaEndpointFactory.datasets())
        yield from fetch_all_pages(request, SigmaPagination)

    def _get_all_files(self) -> Iterator[dict]:
        request = self._get_paginated(endpoint=SigmaEndpointFactory.files())
        yield from fetch_all_pages(request, SigmaPagination)

    def _get_all_members(self) -> Iterator[dict]:
        request = self._get_paginated(endpoint=SigmaEndpointFactory.members())
        yield from fetch_all_pages(request, SigmaPagination)

    def _get_all_workbooks(self) -> Iterator[dict]:
        request = self._get_paginated(endpoint=SigmaEndpointFactory.workbooks())
        yield from fetch_all_pages(request, SigmaPagination)

    def _get_elements_per_page(
        self, page: dict, workbook_id: str
    ) -> Iterator[dict]:
        page_id = page["pageId"]
        request = self._get_paginated(
            SigmaEndpointFactory.elements(workbook_id, page_id)
        )
        elements = fetch_all_pages(request, SigmaPagination)
        for element in elements:
            if element.get("type") not in _DATA_ELEMENTS:
                continue
            yield {
                **element,
                "workbook_id": workbook_id,
                "page_id": page_id,
            }

    def _get_all_elements(self, workbooks: list[dict]) -> Iterator[dict]:
        for workbook in workbooks:
            workbook_id = workbook["workbookId"]

            request = self._get_paginated(
                SigmaEndpointFactory.pages(workbook_id)
            )
            pages = fetch_all_pages(request, SigmaPagination)

            for page in pages:
                yield from self._get_elements_per_page(
                    page=page, workbook_id=workbook_id
                )

    @retry(
        (ConnectionError,),
        max_retries=_RETRY_NUMBER,
        base_ms=_RETRY_BASE_MS,
        log_exc_info=True,
    )
    def _get_lineage(self, lineage_context: LineageContext) -> Lineage:
        """
        return the lineage from API and other ids needed to characterize
        lineage in castor
        """
        workbook_id = lineage_context.workbook_id
        element_id = lineage_context.element_id
        endpoint = SigmaEndpointFactory.lineage(workbook_id, element_id)
        return Lineage(lineage=self._get(endpoint), context=lineage_context)

    @staticmethod
    def _lineage_context(elements: list[dict]) -> list[LineageContext]:
        """
        Helper function to prepare context for lineage retrieval.
        Elements without associated columns are skipped.
        """
        contexts: list[LineageContext] = []
        for element in elements:
            if element.get("columns") is None:
                continue

            context = LineageContext(
                workbook_id=element["workbook_id"],
                element_id=element["elementId"],
            )
            contexts.append(context)
        return contexts

    def _get_all_lineages(self, elements: list[dict]) -> Iterator[dict]:
        """
        The safe mode is temporarily modified to include 403 errors.

        Due to concurrency issues, we force a refresh of the token in hopes that
        the lineage extraction takes less than the token expiration time of
        1 hour.
        """
        safe_mode = self._safe_mode
        self._safe_mode = SIGMA_SAFE_MODE_LINEAGE

        lineage_context = self._lineage_context(elements)

        with ThreadPoolExecutor(max_workers=_THREADS_LINEAGE) as executor:
            results = executor.map(self._get_lineage, lineage_context)

        for lineage in results:
            if not lineage.lineage:
                continue

            yield {
                **lineage.lineage,
                "workbook_id": lineage.context.workbook_id,
                "element_id": lineage.context.element_id,
            }

        self._safe_mode = safe_mode

    @staticmethod
    def _yield_deduplicated_queries(
        queries: Iterable[dict], workbook_id: str
    ) -> Iterator[dict]:
        """
        Returns unique queries for a workbook. This is necessary because the API
        unfortunately returns duplicate entries for some workbook elements.
        """
        seen_elements = set()

        for query in queries:
            element_id = query["elementId"]
            if element_id in seen_elements:
                continue

            seen_elements.add(element_id)
            yield {**query, "workbook_id": workbook_id}

    def _get_all_queries(self, workbooks: list[dict]) -> Iterator[dict]:
        for workbook in workbooks:
            workbook_id = workbook["workbookId"]
            request = self._get_paginated(
                SigmaEndpointFactory.queries(workbook_id),
                limit=SIGMA_QUERIES_PAGINATION_LIMIT,
            )
            queries = fetch_all_pages(request, SigmaPagination)

            yield from self._yield_deduplicated_queries(queries, workbook_id)

    def _get_all_dataset_sources(self, datasets: list[dict]) -> Iterator[dict]:
        yield from SigmaSourcesTransformer(self).get_dataset_sources(datasets)

    def _get_all_workbook_sources(
        self, workbooks: list[dict]
    ) -> Iterator[dict]:
        yield from SigmaSourcesTransformer(self).get_workbook_sources(workbooks)

    def fetch(
        self,
        asset: SigmaAsset,
        datasets: Optional[list[dict]] = None,
        elements: Optional[list[dict]] = None,
        workbooks: Optional[list[dict]] = None,
    ) -> Iterator[dict]:
        """Returns the needed metadata for the queried asset"""
        if asset == SigmaAsset.DATAMODELS:
            yield from self._get_all_datamodels()

        elif asset == SigmaAsset.DATASETS:
            yield from self._get_all_datasets()

        elif asset == SigmaAsset.DATASET_SOURCES:
            if datasets is None:
                raise ValueError("Missing datasets to extract dataset sources")

            yield from self._get_all_dataset_sources(datasets)

        elif asset == SigmaAsset.ELEMENTS:
            if workbooks is None:
                raise ValueError("Missing workbooks to extract elements")

            yield from self._get_all_elements(workbooks)

        elif asset == SigmaAsset.FILES:
            yield from self._get_all_files()

        elif asset == SigmaAsset.LINEAGES:
            if elements is None:
                raise ValueError("Missing elements to extract lineage")

            yield from self._get_all_lineages(elements)

        elif asset == SigmaAsset.MEMBERS:
            yield from self._get_all_members()

        elif asset == SigmaAsset.QUERIES:
            if workbooks is None:
                raise ValueError("Missing workbooks to extract queries")

            yield from self._get_all_queries(workbooks)

        elif asset == SigmaAsset.WORKBOOKS:
            yield from self._get_all_workbooks()

        elif asset == SigmaAsset.WORKBOOK_SOURCES:
            if workbooks is None:
                raise ValueError(
                    "Missing workbooks to extract workbook sources"
                )

            yield from self._get_all_workbook_sources(workbooks)

        else:
            raise ValueError(f"This asset {asset} is unknown")



from collections.abc import Iterator
from functools import partial
from http import HTTPStatus
from typing import Optional

from ....utils import APIClient, BearerAuth, RequestSafeMode, fetch_all_pages
from ..assets import NotionAsset
from .constants import CASTOR_NOTION_USER_AGENT, NOTION_BASE_URL, NOTION_VERSION
from .credentials import NotionCredentials
from .endpoints import NotionEndpointFactory
from .pagination import NotionPagination

VOLUME_IGNORED = 10
IGNORED_ERROR_CODES = (HTTPStatus.BAD_GATEWAY, HTTPStatus.NOT_FOUND)
NOTION_SAFE_MODE = RequestSafeMode(
    max_errors=VOLUME_IGNORED,
    status_codes=IGNORED_ERROR_CODES,
)
NOTION_BASE_HEADERS = {
    "Notion-Version": NOTION_VERSION,
    "User-Agent": CASTOR_NOTION_USER_AGENT,
}

NOTION_DEFAULT_TIMEOUT_S = 180


def _search_filter(asset: str) -> dict[str, dict[str, str]]:
    return {"filter": {"value": asset, "property": "object"}}


class NotionAuth(BearerAuth):
    def __init__(self, token: str):
        self.token = token

    def fetch_token(self):
        return self.token


class NotionClient(APIClient):
    def __init__(
        self,
        credentials: NotionCredentials,
        safe_mode: Optional[RequestSafeMode] = None,
    ):
        auth = NotionAuth(token=credentials.token)
        super().__init__(
            host=NOTION_BASE_URL,
            auth=auth,
            headers=NOTION_BASE_HEADERS,
            safe_mode=safe_mode or NOTION_SAFE_MODE,
            timeout=NOTION_DEFAULT_TIMEOUT_S,
        )

    def users(self) -> Iterator[dict]:
        request = partial(self._get, endpoint=NotionEndpointFactory.users())
        yield from fetch_all_pages(request, NotionPagination)

    def _page_listing(self) -> Iterator[dict]:
        request = partial(
            self._post,
            endpoint=NotionEndpointFactory.search(),
            data=_search_filter("page"),
        )
        yield from fetch_all_pages(request, NotionPagination)

    def _blocks(self, block_id: str) -> Iterator[dict]:
        request = partial(
            self._get,
            endpoint=NotionEndpointFactory.blocks(block_id),
        )
        yield from fetch_all_pages(request, NotionPagination)

    def databases(self) -> Iterator[dict]:
        request = partial(
            self._post,
            endpoint=NotionEndpointFactory.search(),
            data=_search_filter("database"),
        )
        yield from fetch_all_pages(request, NotionPagination)

    def recursive_blocks(self, block_id: str) -> Iterator[dict]:
        """Fetch recursively all children blocks of a given block or page"""
        blocks = self._blocks(block_id)
        for block in blocks:
            if block["has_children"] and block.get("type") != "child_page":
                children = self.recursive_blocks(block["id"])
                block["child_blocks"] = list(children)

            yield block

    def pages(self) -> Iterator[dict]:
        """Fetch all pages with its whole content"""
        for page in self._page_listing():
            if page.get("object") == "database":
                # Notion Search API filter for page doesn't work
                continue
            content = list(self.recursive_blocks(page["id"]))
            page["child_blocks"] = content
            yield page

    def fetch(self, asset: NotionAsset) -> Iterator[dict]:
        """Returns the needed metadata for the queried asset"""
        if asset == NotionAsset.PAGES:
            yield from self.pages()

        elif asset == NotionAsset.DATABASES:
            yield from self.databases()

        elif asset == NotionAsset.USERS:
            yield from self.users()

        else:
            raise ValueError(f"This asset {asset} is unknown")



import logging
from functools import partial
from http import HTTPStatus
from typing import Callable, Optional

from ....utils import (
    APIClient,
    BearerAuth,
    RequestSafeMode,
    SerializedAsset,
    fetch_all_pages,
)
from ..assets import CoalesceAsset, CoalesceQualityAsset
from .credentials import CoalesceCredentials
from .endpoint import (
    CoalesceEndpointFactory,
)
from .pagination import CoalescePagination

logger = logging.getLogger(__name__)


COALESCE_PAGE_SIZE = 300
COALESCE_PAGE_SIZE_RUN_RESULTS = 1_000

COALESCE_TIMEOUT_SECONDS = 90

_MAX_ERRORS = 200

COALESCE_SAFE_MODE = RequestSafeMode(
    status_codes=(HTTPStatus.INTERNAL_SERVER_ERROR,),
    max_errors=_MAX_ERRORS,
)


def _run_result_payload(
    environment_id: str,
    result: dict,
    query_result: dict,
) -> dict:
    return {
        "environment_id": environment_id,
        "node_id": result["nodeID"],
        "node_name": result["name"],
        "test_name": query_result["name"],
        "start_time": query_result["startTime"],
        "end_time": query_result["endTime"],
        "status": query_result["status"],
        "success": query_result["success"],
        "isRunning": query_result["isRunning"],
    }


class CoalesceBearerAuth(BearerAuth):
    """Bearer Authentication for Coalesce"""

    def fetch_token(self) -> Optional[str]:
        pass

    def __init__(self, token: str):
        self._token = token


class CoalesceClient(APIClient):
    """REST API client to extract data from Coalesce"""

    def __init__(
        self,
        credentials: CoalesceCredentials,
    ):
        auth = CoalesceBearerAuth(token=credentials.token)
        super().__init__(
            host=credentials.host,
            auth=auth,
            safe_mode=COALESCE_SAFE_MODE,
            timeout=COALESCE_TIMEOUT_SECONDS,
        )

    def _get_paginated(
        self,
        endpoint: str,
        limit: int = COALESCE_PAGE_SIZE,
        params: Optional[dict] = None,
    ) -> Callable:
        return partial(
            self._get,
            retry_on_timeout=False,  # explained in the docstring
            endpoint=endpoint,
            params={
                "limit": limit,
                **(params or dict()),
            },
        )

    def _fetch_environments(self) -> SerializedAsset:
        endpoint = CoalesceEndpointFactory.environments()
        request = self._get_paginated(endpoint=endpoint)
        result = fetch_all_pages(request, CoalescePagination)
        return list(result)

    def _fetch_env_nodes(self, environment_id: int) -> SerializedAsset:
        endpoint = CoalesceEndpointFactory.nodes(environment_id=environment_id)
        request = self._get_paginated(
            endpoint=endpoint,
            params={"detail": "true"},
        )
        result = fetch_all_pages(request, CoalescePagination)
        return [
            {
                **node,
                "environment_id": environment_id,
            }
            for node in result
        ]

    def _fetch_all_nodes(self) -> SerializedAsset:
        environments = self._fetch_environments()
        total = len(environments)
        nodes: list[dict] = []

        for index, env in enumerate(environments):
            env_id = env["id"]
            logger.info(f"Fetching nodes for env #{env_id} - {index}/{total}")
            nodes.extend(self._fetch_env_nodes(env_id))
            logger.info(f"{len(nodes)} nodes extracted so far")
        return nodes

    def _fetch_runs(self, starting_from: str) -> SerializedAsset:
        endpoint = CoalesceEndpointFactory.runs()
        params = {
            "orderBy": "runEndTime",
            "orderByDirection": "asc",
            "startingFrom": starting_from,
        }
        request = self._get_paginated(
            endpoint=endpoint,
            params=params,
            limit=COALESCE_PAGE_SIZE_RUN_RESULTS,
        )
        return list(fetch_all_pages(request, CoalescePagination))

    def _fetch_run_results(self, run_id: str) -> SerializedAsset:
        endpoint = CoalesceEndpointFactory.run_results(run_id)
        result = self._get(endpoint=endpoint)
        return result["data"]

    def _run_results_by_run(
        self,
        environment_id: str,
        run_id: str,
    ) -> SerializedAsset:
        run_results: list[dict] = []
        for result in self._fetch_run_results(run_id):
            for query_result in result["queryResults"]:
                if query_result["type"] != "sqlTest":
                    continue
                run_result = _run_result_payload(
                    environment_id,
                    result,
                    query_result,
                )
                run_results.append(run_result)
        return run_results

    def _fetch_all_run_results(
        self,
        starting_from: str,
    ) -> SerializedAsset:
        run_results: list[dict] = []

        runs = self._fetch_runs(starting_from)
        total = len(runs)

        for index, run in enumerate(runs):
            logger.info(f"Extracting run results ({index}/{total})")
            run_id = run["id"]
            environment_id = run["environmentID"]
            current_results = self._run_results_by_run(environment_id, run_id)
            run_results.extend(current_results)
        return run_results

    def fetch(
        self, asset: CoalesceAsset, starting_from=None
    ) -> SerializedAsset:
        """Extract the given Coalesce Asset"""
        if asset in (CoalesceAsset.NODES, CoalesceQualityAsset.NODES):
            return self._fetch_all_nodes()
        elif asset == CoalesceQualityAsset.RUN_RESULTS:
            return self._fetch_all_run_results(starting_from=starting_from)
        raise AssertionError(
            f"Asset {asset} is not supported by CoalesceClient"
        )



from argparse import Namespace

from .argument_parser import parse_filled_arguments


class MockArgumentParser:
    def __init__(self):
        self.attributes = {}

    def add_argument(self, name, value):
        self.attributes[name] = value

    def parse_args(self) -> Namespace:
        return Namespace(**self.attributes)


def test_parse_filled_arguments():
    parser = MockArgumentParser()
    parser.add_argument("filled", "value")
    parser.add_argument("unfilled", None)
    parser.add_argument("empty_str", "")

    expected = {"filled": "value", "empty_str": ""}
    assert parse_filled_arguments(parser) == expected



import logging
from http import HTTPStatus
from typing import TYPE_CHECKING, Callable, Iterator

from ....utils import retry_request
from .endpoints import SigmaEndpointFactory

if TYPE_CHECKING:
    from .client import SigmaClient

logger = logging.getLogger(__name__)

SIGMA_CONNECTION_PATH_MAX_RETRY = 1
SIGMA_CONNECTION_PATH_SLEEP_MS = 30_000  # 30 seconds


class SigmaSourcesTransformer:
    """Retrieves asset sources and enhances them with additional information."""

    def __init__(self, api_client: "SigmaClient"):
        self.api_client = api_client

    @retry_request(
        status_codes=(HTTPStatus.TOO_MANY_REQUESTS,),
        max_retries=SIGMA_CONNECTION_PATH_MAX_RETRY,
        base_ms=SIGMA_CONNECTION_PATH_SLEEP_MS,
    )
    def _get_connection_path(self, table_id: str) -> dict:
        """Retrieves the connection path for a given table id"""
        return self.api_client._get(
            endpoint=SigmaEndpointFactory.connection_path(table_id)
        )

    def _map_table_id_to_connection_path(
        self, all_sources: list
    ) -> dict[str, dict]:
        """Maps a table id to its connection and path information."""
        logger.info("Mapping table ids to connection and path information")

        unique_table_ids = {
            source["inodeId"]
            for asset_sources in all_sources
            for source in asset_sources["sources"]
            if source["type"] == "table"
        }

        return {
            table_id: self._get_connection_path(table_id)
            for table_id in unique_table_ids
        }

    @staticmethod
    def _enhance_table_source(source: dict, table_to_path: dict) -> dict:
        """
        Combines a single table source with its connection and path information.
        """
        if source["type"] != "table":
            return source

        path_info = table_to_path.get(source["inodeId"], {})
        source["connectionId"] = path_info.get("connectionId")
        source["path"] = path_info.get("path")
        return source

    def _transform_sources(
        self, all_sources: list, table_to_path: dict
    ) -> Iterator[dict]:
        """
        Yields all sources, with table sources being enhanced with additional information.
        """
        logger.info("Merging sources with table information")

        for asset_sources in all_sources:
            enhanced_sources = [
                self._enhance_table_source(source, table_to_path)
                for source in asset_sources["sources"]
            ]

            yield {
                "asset_id": asset_sources["asset_id"],
                "sources": enhanced_sources,
            }

    def _get_all_sources(
        self, endpoint: Callable[[str], str], asset_ids: set[str]
    ) -> Iterator[dict]:
        """Returns transformed sources for the given assets"""
        all_sources = []

        for asset_id in asset_ids:
            sources = self.api_client._get(endpoint=endpoint(asset_id))
            all_sources.append({"asset_id": asset_id, "sources": sources})

        table_to_path = self._map_table_id_to_connection_path(all_sources)

        yield from self._transform_sources(all_sources, table_to_path)

    def get_dataset_sources(self, datasets: list[dict]) -> Iterator[dict]:
        asset_ids = {dataset["datasetId"] for dataset in datasets}
        yield from self._get_all_sources(
            endpoint=SigmaEndpointFactory.dataset_sources, asset_ids=asset_ids
        )

    def get_workbook_sources(self, workbooks: list[dict]) -> Iterator[dict]:
        asset_ids = {workbook["workbookId"] for workbook in workbooks}
        yield from self._get_all_sources(
            endpoint=SigmaEndpointFactory.workbook_sources, asset_ids=asset_ids
        )



from typing import Any, Optional

from pydantic import Field, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

STRATEGY_ENV_PREFIX = "CATALOG_STRATEGY_"


class StrategyCredentials(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix=STRATEGY_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    base_url: str
    password: str = Field(repr=False)
    username: str

    project_ids: Optional[list[str]] = None

    @field_validator("project_ids", mode="before")
    @classmethod
    def _check_project_ids(cls, project_ids: Any) -> Optional[list[str]]:
        """
        The project IDs are optional and can be either a list of strings
        or single string with project IDs separated by commas.
        """
        if project_ids is None:
            return None

        if isinstance(project_ids, str):
            return [item.strip() for item in project_ids.split(",")]

        if isinstance(project_ids, list):
            return project_ids

        raise ValueError(f"Unexpected type for project_id: {type(project_ids)}")



import logging
from argparse import ArgumentParser

from castor_extractor.utils import parse_filled_arguments  # type: ignore
from castor_extractor.visualization import domo  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    # mandatory
    parser.add_argument(
        "-b",
        "--base-url",
        help="Host name where the server is running",
    )
    parser.add_argument(
        "-a",
        "--api-token",
        help="Token to connect to Domo",
    )
    parser.add_argument(
        "-d",
        "--developer-token",
        help="Developer token to use private API",
    )
    parser.add_argument(
        "-c",
        "--client-id",
        help="Client id of the Domo account",
    )
    parser.add_argument(
        "-C",
        "--cloud-id",
        help="Cloud id is the external warehouse id within Domo",
    )

    parser.add_argument("-o", "--output", help="Directory to write to")

    domo.extract_all(**parse_filled_arguments(parser))



import logging
from collections.abc import Iterable, Iterator
from typing import Union

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    write_json,
    write_summary,
)
from .assets import ThoughtspotAsset
from .client import (
    ThoughtspotClient,
    ThoughtspotCredentials,
)

logger = logging.getLogger(__name__)


def iterate_all_data(
    client: ThoughtspotClient,
) -> Iterable[tuple[ThoughtspotAsset, Union[list, Iterator, dict]]]:
    """Iterate over the extracted data from Thoughtspot"""

    for asset in ThoughtspotAsset:
        logger.info(f"Extracting {asset.value} from API")
        data = client.fetch(asset)
        yield asset, deep_serialize(data)


def extract_all(**kwargs) -> None:
    """
    Extract data from Thoughtspot API
    Store the output files locally under the given output_directory
    """
    _output_directory = kwargs.get("output") or from_env(OUTPUT_DIR)

    credentials = ThoughtspotCredentials(**kwargs)
    client = ThoughtspotClient(credentials=credentials)

    ts = current_timestamp()

    for key, data in iterate_all_data(client):
        filename = get_output_filename(key.name.lower(), _output_directory, ts)
        write_json(filename, list(data))

    write_summary(_output_directory, ts, host=credentials.base_url)



import logging
from argparse import ArgumentParser

from castor_extractor.warehouse import sqlserver  # type:ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    parser.add_argument("-H", "--host", help="MSSQL Host")
    parser.add_argument("-P", "--port", help="MSSQL Port")
    parser.add_argument("-d", "--database", help="MSSQL Database")
    parser.add_argument("-u", "--user", help="MSSQL User")
    parser.add_argument("-p", "--password", help="MSSQL Password")

    parser.add_argument(
        "-s",
        "--skip-queries",
        dest="skip_queries",
        action="store_true",
        help="Skip the extraction of SQL queries",
    )

    parser.add_argument("-o", "--output", help="Directory to write to")

    parser.add_argument(
        "--skip-existing",
        dest="skip_existing",
        action="store_true",
        help="Skips files already extracted instead of replacing them",
    )
    parser.add_argument(
        "--db-allowed",
        nargs="*",
        help="List of databases that should be extracted",
    )
    parser.add_argument(
        "--db-blocked",
        nargs="*",
        help="List of databases that should not be extracted",
    )
    parser.set_defaults(skip_existing=False)

    args = parser.parse_args()

    sqlserver.extract_all(
        host=args.host,
        database=args.database,
        port=args.port,
        user=args.user,
        password=args.password,
        output_directory=args.output,
        skip_existing=args.skip_existing,
        db_allowed=args.db_allowed,
        db_blocked=args.db_blocked,
    )



from .client import SigmaClient


def test_SigmaClient__yield_deduplicated_queries():
    workbook_id = "workbook1"
    mock_queries = [
        {"elementId": "element1", "name": "Query 1"},
        {"elementId": "element2", "name": "Query 2"},
        {"elementId": "element1", "name": "Query 1"},  # Duplicate
        {"elementId": "element3", "name": "Query 3"},
    ]

    queries = list(
        SigmaClient._yield_deduplicated_queries(mock_queries, workbook_id)
    )

    assert len(queries) == 3
    for query in queries:
        assert query["workbook_id"] == workbook_id



from ..assets import SalesforceReportingAsset

queries: dict[SalesforceReportingAsset, str] = {
    SalesforceReportingAsset.DASHBOARDS: """
        SELECT
            CreatedBy.Id,
            CreatedDate,
            Description,
            DeveloperName,
            FolderId,
            FolderName,
            Id,
            IsDeleted,
            LastReferencedDate,
            LastViewedDate,
            NamespacePrefix,
            RunningUserId,
            Title,
            Type
        FROM Dashboard
        WHERE IsDeleted = FALSE
    """,
    SalesforceReportingAsset.FOLDERS: """
    SELECT
        DeveloperName,
        Id,
        Name,
        NamespacePrefix,
        ParentId,
        Type
    FROM Folder
    WHERE Type IN ('Dashboard', 'Report')
    """,
    SalesforceReportingAsset.REPORTS: """
        SELECT
            CreatedBy.Id,
            Description,
            DeveloperName,
            FolderName,
            Format,
            Id,
            IsDeleted,
            LastReferencedDate,
            LastRunDate,
            LastViewedDate,
            Name,
            NamespacePrefix,
            OwnerId
        FROM Report
        WHERE IsDeleted = FALSE
    """,
    SalesforceReportingAsset.USERS: """
        SELECT
            Id,
            Email,
            FirstName,
            LastName,
            CreatedDate
        FROM User
        WHERE UserType = 'Standard'
    """,
    SalesforceReportingAsset.DASHBOARD_COMPONENTS: """
        SELECT
            CustomReportId,
            DashboardId,
            Name
        FROM DashboardComponent
    """,
}



from looker_sdk.rtl.api_settings import SettingsConfig
from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

from ..constants import DEFAULT_LOOKER_TIMEOUT_SECOND, LOOKER_ENV_PREFIX

KEY_LOOKER_TIMEOUT_SECOND = f"{LOOKER_ENV_PREFIX}TIMEOUT_SECOND"


class LookerCredentials(BaseSettings):
    """ValueObject for the credentials"""

    model_config = SettingsConfigDict(
        env_prefix=LOOKER_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    base_url: str
    client_id: str
    client_secret: str = Field(repr=False)
    timeout: int = Field(
        validation_alias=KEY_LOOKER_TIMEOUT_SECOND,
        default=DEFAULT_LOOKER_TIMEOUT_SECOND,
    )

    def to_settings_config(self) -> SettingsConfig:
        return SettingsConfig(
            base_url=self.base_url,
            client_id=self.client_id,
            client_secret=self.client_secret,
            timeout=str(self.timeout),
        )



from functools import partial
from unittest.mock import patch

from ....utils import fetch_all_pages
from .client import NotionClient
from .credentials import NotionCredentials
from .pagination import NotionPagination

MOCK_PAGINATED_RESPONSE = {
    "results": [{"result_id": 1}],
    "next_cursor": "2",
    "has_more": True,
}
MOCK_PAGINATED_RESPONSE_2 = {
    "results": [{"result_id": 2}],
    "next_cursor": None,
    "has_more": False,
}
MOCK_RESPONSE = {"result_id": 3}


@patch.object(NotionClient, "_get")
def test_NotionClient__request(mock_call):
    mock_call.side_effect = [
        MOCK_PAGINATED_RESPONSE,
        MOCK_PAGINATED_RESPONSE_2,
        MOCK_RESPONSE,
    ]

    client = NotionClient(NotionCredentials(token="MockToken"))
    get_call = partial(client._get, endpoint="FAKE")

    response = list(fetch_all_pages(get_call, NotionPagination))
    assert response == [{"result_id": 1}, {"result_id": 2}]

    response = get_call()
    assert response == {"result_id": 3}


MOCK_BLOCK = [{"object": "block", "id": "1", "has_children": True}]
MOCK_BLOCK_CHILDREN = [
    {"object": "block", "id": "2", "has_children": False},
    {"object": "block", "id": "3", "has_children": True, "type": "child_page"},
]


@patch.object(NotionClient, "_blocks")
def test_NotionClient__recursive_blocks(mock_blocks):
    mock_blocks.side_effect = [
        MOCK_BLOCK,
        MOCK_BLOCK_CHILDREN,
    ]

    client = NotionClient(NotionCredentials(token="MockToken"))
    response = list(client.recursive_blocks("1"))

    assert response == [
        {
            "object": "block",
            "id": "1",
            "has_children": True,
            "child_blocks": [
                {"object": "block", "id": "2", "has_children": False},
                {
                    "object": "block",
                    "id": "3",
                    "has_children": True,
                    "type": "child_page",
                },
            ],
        }
    ]



import pytest

from .extract import _validate_user_emails


def test__validate_user_emails():
    with pytest.raises(TypeError):
        _validate_user_emails("toto@tata.com")

    with pytest.raises(TypeError):
        _validate_user_emails({"not": "the", "right": "format"})

    with pytest.raises(ValueError):
        _validate_user_emails([])

    with pytest.raises(TypeError):
        _validate_user_emails([1, 2, 3, 4])

    _validate_user_emails(["admin@toto.com", "tata@toto.com"])



class SodaEndpointFactory:
    """Wrapper class around all endpoints we're using"""

    @classmethod
    def datasets(cls) -> str:
        return "datasets"

    @classmethod
    def checks(cls) -> str:
        return "checks"



from typing import Optional

from .auth import BasicAuth, BearerAuth, CustomAuth


class _MockRequest:
    def __init__(self):
        self.headers = {}


class _CustomAuth(CustomAuth):
    def _authentication_header(self) -> dict[str, str]:
        return {"custom-token": "token"}


class _BearAuth(BearerAuth):
    def fetch_token(self) -> Optional[str]:
        return "token"


def test_BasicAuth():
    prepared_request = _MockRequest()
    auth = BasicAuth(username="simple", password="basic")
    auth.__call__(prepared_request)
    assert prepared_request.headers == {
        "Authorization": "Basic c2ltcGxlOmJhc2lj"
    }


def test_BearerAuth():
    prepared_request = _MockRequest()
    auth = _BearAuth()
    auth.__call__(prepared_request)
    assert prepared_request.headers == {"Authorization": "Bearer token"}

    auth._token = "expired_token"
    auth.__call__(prepared_request)
    assert prepared_request.headers == {"Authorization": "Bearer expired_token"}

    auth.refresh_token()
    auth.__call__(prepared_request)
    assert prepared_request.headers == {"Authorization": "Bearer token"}


def test_CustomAuth():
    prepared_request = _MockRequest()
    auth = _CustomAuth()
    auth.__call__(prepared_request)
    assert prepared_request.headers == {"custom-token": "token"}



from .argument_parser import parse_filled_arguments
from .batch import batch_of_length
from .client import (
    AbstractSourceClient,
    APIClient,
    Auth,
    BasicAuth,
    BearerAuth,
    CustomAuth,
    ExtractionQuery,
    FetchNextPageBy,
    PaginationModel,
    PostgresClient,
    RequestSafeMode,
    ResponseJson,
    SqlalchemyClient,
    build_url,
    fetch_all_pages,
    handle_response,
    uri_encode,
)
from .collection import (
    deduplicate,
    empty_iterator,
    filter_items,
    group_by,
    mapping_from_rows,
)
from .constants import ENCODING_UTF8, OUTPUT_DIR
from .deprecate import deprecate_python
from .env import from_env
from .files import explode, search_files
from .formatter import to_string_array
from .json_stream_write import StreamableList
from .load import load_file
from .object import deep_serialize, getproperty
from .pager import (
    Pager,
    PagerLogger,
    PagerOnId,
    PagerOnIdLogger,
    PagerStopStrategy,
)
from .retry import RetryStrategy, retry, retry_request
from .safe import SafeMode, safe_mode
from .store import AbstractStorage, LocalStorage
from .string import decode_when_bytes, string_to_tuple
from .time import (
    at_midnight,
    current_date,
    current_datetime,
    current_timestamp,
    date_after,
    format_date,
    format_rfc_3339_date,
    past_date,
    timestamp_ms,
    yesterday,
)
from .type import Callback, Getter, JsonType, SerializedAsset
from .url import add_path as add_path_to_url, url_from
from .validation import clean_path, validate_baseurl
from .write import (
    get_output_filename,
    get_summary_filename,
    get_summary_payload,
    write_errors_logs,
    write_json,
    write_summary,
)



from ..assets import QlikAsset

APP_EXTERNAL_ID_KEY = "resourceId"

# API settings
REST_API_BASE_PATH = "api/v1/"

# requests retry settings
RETRY_COUNTS = 3
RETRY_STATUSES = (404, 500, 502, 503, 504)
RETRY_BACKOFF_FACTOR = 0.5

_RESPONSE_BASE_EXPECTED_MSG = "Expecting HTTP response to be a {type}"
RESPONSE_DICT_EXPECTED_MSG = _RESPONSE_BASE_EXPECTED_MSG.format(type="dict")
RESPONSE_LIST_EXPECTED_MSG = _RESPONSE_BASE_EXPECTED_MSG.format(type="list")


ASSET_PATHS: dict[QlikAsset, str] = {
    QlikAsset.CONNECTIONS: "data-connections",
    QlikAsset.SPACES: "spaces",
    QlikAsset.USERS: "users",
    QlikAsset.APPS: "items?resourceType=app",
    QlikAsset.LINEAGE: "apps/{app_id}/data/lineage",
}

SCOPED_ASSETS = (QlikAsset.LINEAGE, QlikAsset.MEASURES)



from contextlib import contextmanager
from functools import partial
from http import HTTPStatus
from typing import Iterator, Optional

from google.auth.transport.requests import Request
from google.oauth2.service_account import Credentials

from ....utils import (
    APIClient,
    BearerAuth,
    RequestSafeMode,
    fetch_all_pages,
)
from .credentials import LookerStudioCredentials
from .endpoints import LookerStudioAPIEndpoint
from .enums import LookerStudioAssetType
from .pagination import LookerStudioPagination


@contextmanager
def temporary_safe_request(client: APIClient):
    """
    Allows applying the safe mode only to specific requests instead of all
    API calls.
    """
    client._safe_mode = RequestSafeMode(
        max_errors=float("inf"),
        status_codes=(HTTPStatus.UNAUTHORIZED,),
    )
    try:
        yield client
    finally:
        client._safe_mode = RequestSafeMode()


class LookerStudioAPIAuth(BearerAuth):
    def __init__(
        self,
        credentials: LookerStudioCredentials,
        subject: Optional[str] = None,
    ):
        """
        Instantiates the service account credentials.
        If a `subject` email is passed, the service account will impersonate
        that user and make requests on that user's behalf.
        """
        self._credentials = Credentials.from_service_account_info(
            credentials.model_dump(),
            scopes=credentials.scopes,
        )
        if subject:
            self._credentials = self._credentials.with_subject(subject)

    def fetch_token(self):
        self._credentials.refresh(Request())
        return self._credentials.token


class LookerStudioAPIClient(APIClient):
    def __init__(self, credentials: LookerStudioCredentials):
        auth = LookerStudioAPIAuth(credentials=credentials)
        super().__init__(auth=auth)

        self._credentials = credentials

    def _is_private_asset(self, asset_name: str) -> bool:
        """
        Returns True if the asset is not viewable by anyone other than the owner.

        The permissions	dict contains `Role: Member[]` key-value pairs and has
        at least one key-value pair to define the asset's unique OWNER.
        If another key is present, it means the asset was shared with
        another person or group.

        See also https://developers.google.com/looker-studio/integrate/api/reference/types#Permissions
        """
        with temporary_safe_request(self):
            data = self._get(LookerStudioAPIEndpoint.permissions(asset_name))
            permissions = data.get("permissions")

            if not permissions:
                return True

            return len(permissions.keys()) == 1

    def _user_assets(
        self, asset_type: LookerStudioAssetType, user_email: str
    ) -> Iterator[dict]:
        """
        Yields all assets of the given type, owned by the given user and visible
        by other members.
        """
        request = partial(
            self._get,
            LookerStudioAPIEndpoint.search(),
            params={"assetTypes": [asset_type.value]},
        )
        assets = fetch_all_pages(request, LookerStudioPagination)

        for asset in assets:
            asset_name = asset["name"]
            owner = asset.get("owner", "")
            if owner == user_email and not self._is_private_asset(asset_name):
                yield asset

    def _impersonate_user(self, user_email: str):
        self._auth = LookerStudioAPIAuth(
            credentials=self._credentials, subject=user_email
        )

    def fetch_user_assets(self, user_email: str) -> Iterator[dict]:
        """Yields assets (reports and data sources) shared by the given user."""
        self._impersonate_user(user_email)

        reports = self._user_assets(
            asset_type=LookerStudioAssetType.REPORT,
            user_email=user_email,
        )
        data_sources = self._user_assets(
            asset_type=LookerStudioAssetType.DATA_SOURCE,
            user_email=user_email,
        )

        yield from reports
        yield from data_sources



from .format import (
    _HAS_DUPLICATE_KEY,
    SalesforceFormatter,
    _detect_duplicate_labels,
    _field_description,
    _name,
    _remove_duplicates,
)


def _tables_sobjects() -> tuple[dict[str, str], ...]:
    """Returns 4 sobjects with 2 sharing the same label"""
    a = {"Label": "a", "QualifiedApiName": "a_one"}
    b = {"Label": "b", "QualifiedApiName": "b"}
    c = {"Label": "c", "QualifiedApiName": "c_unique_so_doesnt_matter"}
    a_prime = {"Label": "a", "QualifiedApiName": "a_two"}
    b_exact_duplicate = {"Label": "b", "QualifiedApiName": "b"}
    return a, b, c, a_prime, b_exact_duplicate


def _columns_sobjects() -> dict[str, list[dict]]:
    a = {"Label": "First Name", "QualifiedApiName": "owner_name"}
    b = {"Label": "First Name", "QualifiedApiName": "editor_name"}
    c = {"Label": "Foo Bar", "QualifiedApiName": "foo_bar"}
    return {
        "table_1": [a, b],
        "table_2": [c],
    }


def test__field_description():
    field = {}
    assert _field_description(field) == ""

    definition = {}
    field = {"FieldDefinition": definition}
    assert _field_description(field) == ""

    definition.update({"Description": "foo"})
    assert "foo" in _field_description(field)

    field.update({"InlineHelpText": "bar"})
    assert "bar" in _field_description(field)

    definition.update({"ComplianceGroup": "bim"})
    assert "bim" in _field_description(field)

    definition.update({"SecurityClassification": "bam"})
    description = _field_description(field)

    assert "bam" in description
    expected = (
        "- Description: foo\n"
        "- Help Text: bar\n"
        "- Compliance Categorization: bim\n"
        "- Data Sensitivity Level: bam"
    )
    assert description == expected


def test__name():
    unique_sobject = {
        "Label": "First Name",
        "QualifiedApiName": "first_name",
        _HAS_DUPLICATE_KEY: False,
    }
    assert _name(unique_sobject) == "First Name"

    duplicate_sobject = {
        "Label": "First Name",
        "QualifiedApiName": "first_name",
        _HAS_DUPLICATE_KEY: True,
    }
    assert _name(duplicate_sobject) == "First Name (first_name)"

    empty_label_sobject = {
        "Label": "",
        "QualifiedApiName": "empty_label",
        _HAS_DUPLICATE_KEY: False,
    }
    assert _name(empty_label_sobject) == "empty_label"


def test__detect_duplicate_labels():
    objects = [
        {"Label": "Foo"},
        {"Label": "Bar"},
        {"Label": "Foo"},
    ]

    objects = _detect_duplicate_labels(objects)
    assert objects == [
        {"Label": "Foo", _HAS_DUPLICATE_KEY: True},
        {"Label": "Bar", _HAS_DUPLICATE_KEY: False},
        {"Label": "Foo", _HAS_DUPLICATE_KEY: True},
    ]


def test__remove_duplicates():
    objects = [
        {"QualifiedApiName": "Foo"},
        {"QualifiedApiName": "Bar"},
        {"QualifiedApiName": "Foo"},
    ]

    objects = _remove_duplicates(objects)
    assert len(objects) == 2
    names = {sobject["QualifiedApiName"] for sobject in objects}
    assert names == {"Foo", "Bar"}


def test_salesforce_formatter_tables():
    sobjects = [*_tables_sobjects()]
    tables = [t for t in SalesforceFormatter.tables(sobjects)]
    expected_names = {"a (a_one)", "a (a_two)", "b", "c"}
    payload_names = {t["table_name"] for t in tables}
    assert len(tables) == 4  # we only keep one "b"
    assert payload_names == expected_names


def test_salesforce_formatter_columns():
    sobjects = _columns_sobjects()
    columns = SalesforceFormatter.columns(sobjects)
    column_ids = {c["id"] for c in columns}
    expected_column_ids = {
        "table_1.First Name (owner_name)",
        "table_1.First Name (editor_name)",
        "table_2.Foo Bar",
    }
    assert column_ids == expected_column_ids



import pytest

from .extract import (
    REDSHIFT_SERVERLESS,
    _query_builder,
)


@pytest.mark.parametrize(
    "serverless_param,env_param,expected",
    [
        (True, "False", True),
        (False, "True", True),
        (None, "TRUE", True),
        (None, "TrUe", True),
        (None, "FAlSE", False),
        (None, "False", False),
        (None, None, False),
        (True, None, True),
    ],
)
def test__query_builder(serverless_param, env_param, expected, monkeypatch):
    params = {"serverless": serverless_param}
    monkeypatch.setenv(REDSHIFT_SERVERLESS, env_param)

    assert _query_builder(params).is_serverless == expected



import logging
from argparse import ArgumentParser

from castor_extractor.utils import parse_filled_arguments  # type: ignore
from castor_extractor.visualization import metabase  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    # mandatory
    parser.add_argument(
        "-H",
        "--host",
        help="Host name where the server is running",
    )
    parser.add_argument("-P", "--port", help="TCP/IP port number")
    parser.add_argument("-d", "--database", help="Database name")
    parser.add_argument("-s", "--schema", help="Schema name")
    parser.add_argument("-u", "--user", help="Username")
    parser.add_argument("-p", "--password", help="Password")
    parser.add_argument(
        "-k",
        "--encryption_secret_key",
        help="Encryption secret key",
    )
    parser.add_argument(
        "--require_ssl",
        action="store_true",
        help="Require SSL",
    )

    parser.add_argument("-o", "--output", help="Directory to write to")

    args = parse_filled_arguments(parser)
    credentials = metabase.MetabaseDbCredentials(**args)

    client = metabase.DbClient(credentials)

    metabase.extract_all(client, **args)



from argparse import ArgumentParser

from castor_extractor.utils import parse_filled_arguments  # type: ignore
from castor_extractor.visualization import looker_studio  # type: ignore


def main():
    parser = ArgumentParser()
    parser.add_argument(
        "-c",
        "--credentials",
        help="File path to Service Account credentials with Looker Studio access",
    )
    parser.add_argument(
        "-a",
        "--admin-email",
        help="Email of a Google Workspace user with admin access",
    )
    parser.add_argument(
        "-b",
        "--bigquery-credentials",
        help=(
            "Optional: file path to Service Account credentials with BigQuery access. "
            "This can be the same file path as for Looker Studio."
        ),
    )
    parser.add_argument(
        "--skip-view-activity-logs",
        action="store_true",
        default=False,
        help="Skips the extraction of activity logs",
    )
    parser.add_argument(
        "--users-file-path",
        help=(
            "Optional path to a JSON file with user email addresses "
            'as a list of strings (e.g. ["foo@bar.com", "fee@bar.com"]). '
            "If provided, only extracts assets owned by the specified users."
        ),
    )

    parser.add_argument("-o", "--output", help="Directory to write to")

    looker_studio.extract_all(**parse_filled_arguments(parser))



from typing import Optional

from pydantic import ConfigDict
from pydantic.alias_generators import to_camel

from ....utils import (
    FetchNextPageBy,
    PaginationModel,
)


class PowerBiPagination(PaginationModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    fetch_by: FetchNextPageBy = FetchNextPageBy.URL

    activity_event_entities: list
    continuation_uri: Optional[str] = None
    last_result_set: bool = False

    def is_last(self) -> bool:
        return self.last_result_set

    def next_page_payload(self) -> Optional[str]:
        return self.continuation_uri

    def page_results(self) -> list:
        return self.activity_event_entities



from datetime import date
from unittest.mock import Mock, call, patch

import pytest

from .authentication import msal
from .client import PowerbiClient
from .constants import Keys
from .credentials import CLIENT_APP_BASE, REST_API_BASE_PATH, PowerbiCredentials
from .endpoints import PowerBiEndpointFactory

FAKE_TENANT_ID = "IamFake"
FAKE_CLIENT_ID = "MeTwo"
FAKE_SECRET = "MeThree"

ENDPOINT_FACTORY = PowerBiEndpointFactory(
    login_url=CLIENT_APP_BASE,
    api_base=REST_API_BASE_PATH + "/",  # superfluous "/" to test resiliency
)


@pytest.fixture
def mock_msal():
    with patch.object(msal, "ConfidentialClientApplication") as mock_app:
        mock_app.return_value.acquire_token_for_client.return_value = {
            "access_token": "fake_token"
        }
        yield mock_app


@pytest.fixture
def power_bi_client(mock_msal):
    creds = PowerbiCredentials(
        tenant_id=FAKE_TENANT_ID,
        client_id=FAKE_CLIENT_ID,
        secret=FAKE_SECRET,
    )
    return PowerbiClient(creds)


def test__access_token(power_bi_client, mock_msal):
    # Valid token scenario
    valid_token = "mock_token"
    mock_response = {"access_token": valid_token}
    returning_valid_token = Mock(return_value=mock_response)
    mock_msal.return_value.acquire_token_for_client = returning_valid_token

    assert power_bi_client._auth.fetch_token() == valid_token

    # Invalid token scenario
    invalid_response = {"not_access_token": "666"}
    returning_invalid_token = Mock(return_value=invalid_response)
    mock_msal.return_value.acquire_token_for_client = returning_invalid_token

    with pytest.raises(ValueError):
        power_bi_client._auth.fetch_token()


def test__datasets(power_bi_client):
    with patch.object(power_bi_client, "_get") as mocked_get:
        mocked_get.return_value = {"value": [{"id": 1, "type": "dataset"}]}
        datasets = list(power_bi_client._datasets())
        mocked_get.assert_called_with(ENDPOINT_FACTORY.datasets())
        assert datasets == [{"id": 1, "type": "dataset"}]


def test__dashboards(power_bi_client):
    with patch.object(power_bi_client, "_get") as mocked_get:
        mocked_get.return_value = {"value": [{"id": 1, "type": "dashboard"}]}
        dashboards = list(power_bi_client._dashboards())
        mocked_get.assert_called_with(ENDPOINT_FACTORY.dashboards())
        assert dashboards == [{"id": 1, "type": "dashboard"}]


def test__reports(power_bi_client):
    with patch.object(power_bi_client, "_get") as mocked_get:
        mocked_get.side_effect = [
            {"value": [{"id": 1, "type": "report"}]},
            {
                "value": [
                    {"name": "page_name", "displayName": "page", "order": 0}
                ]
            },
        ]
        reports = list(power_bi_client._reports())
        calls = [
            call(ENDPOINT_FACTORY.reports()),
            call(ENDPOINT_FACTORY.pages("1"), retry_on_timeout=False),
        ]
        mocked_get.assert_has_calls(calls)
        assert reports == [
            {
                "id": 1,
                "type": "report",
                "pages": [
                    {"name": "page_name", "displayName": "page", "order": 0}
                ],
            }
        ]


def test__workspace_ids(power_bi_client):
    with patch.object(power_bi_client, "_get") as mocked_get:
        mocked_get.return_value = [{"id": 1000}, {"id": 1001}, {"id": 1003}]

        ids = power_bi_client._workspace_ids()
        assert ids == [1000, 1001, 1003]

        params = {
            Keys.INACTIVE_WORKSPACES: True,
            Keys.PERSONAL_WORKSPACES: True,
        }

        mocked_get.assert_called_with(
            ENDPOINT_FACTORY.workspace_ids(),
            params=params,
        )


@patch.object(PowerbiClient, "_get_scan_result")
@patch.object(PowerbiClient, "_wait_for_scan_result")
@patch.object(PowerbiClient, "_create_scan")
@patch.object(PowerbiClient, "_workspace_ids")
def test__metadata(
    mock_workspace_ids,
    mock_create_scan,
    mock_wait_for_scan,
    mock_get_scan_result,
    power_bi_client,
):
    mock_workspace_ids.return_value = list(range(200))
    mock_create_scan.return_value = 314
    mock_wait_for_scan.return_value = True
    mock_get_scan_result.return_value = [{"workspace_id": 1871}]

    result = list(power_bi_client._metadata())

    assert result == [{"workspace_id": 1871}, {"workspace_id": 1871}]


def test__activity_events(power_bi_client):
    day = date.today()
    mocked_get_results = [
        {
            Keys.ACTIVITY_EVENT_ENTITIES: [
                {
                    "id": "foo",
                    "name": "Foo",
                    "ClientIP": "1.1.1.1",
                    "UserAgent": "Mozilla/5.0",
                },
                {
                    "id": "bar",
                    "name": "Bar",
                    "ClientIP": "1.1.1.2",
                    "UserAgent": "Mozilla/5.0",
                },
            ],
            Keys.LAST_RESULT_SET: False,
            Keys.CONTINUATION_URI: "https://next-call-1",
        },
        {
            Keys.ACTIVITY_EVENT_ENTITIES: [
                {
                    "id": "baz",
                    "name": "Baz",
                    "ClientIP": "1.1.1.3",
                    "UserAgent": "Mozilla/5.0",
                }
            ],
            Keys.LAST_RESULT_SET: False,
            Keys.CONTINUATION_URI: "https://next-call-2",
        },
        {
            Keys.ACTIVITY_EVENT_ENTITIES: [
                {
                    "id": "biz",
                    "name": "Biz",
                    "ClientIP": "1.1.1.4",
                    "UserAgent": "Mozilla/5.0",
                }
            ],
            Keys.LAST_RESULT_SET: True,
            Keys.CONTINUATION_URI: None,
        },
    ]

    with patch.object(power_bi_client, "_get") as mocked_get:
        mocked_get.side_effect = mocked_get_results

        result = list(power_bi_client._activity_events(day=day))
        assert result == [
            {"id": "foo", "name": "Foo"},
            {"id": "bar", "name": "Bar"},
            {"id": "baz", "name": "Baz"},
            {"id": "biz", "name": "Biz"},
        ]

        expected_calls = [
            call(endpoint=ENDPOINT_FACTORY.activity_events(day=day)),
            call(endpoint="https://next-call-1"),
            call(endpoint="https://next-call-2"),
        ]
        mocked_get.assert_has_calls(expected_calls)


def test_test_connection(power_bi_client):
    with patch.object(power_bi_client._auth, "refresh_token") as mock_refresh:
        power_bi_client.test_connection()
        mock_refresh.assert_called_once()



import logging
import os
from argparse import ArgumentParser

from castor_extractor import file_checker  # type: ignore
from castor_extractor.utils import (  # type: ignore
    LocalStorage,  # type: ignore
    explode,
    search_files,
)

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")
logger = logging.getLogger(__name__)

WarehouseTemplate = file_checker.GenericWarehouseFileTemplate

Ids = set[str]
_ID_KEY = "id"


def process(directory: str, verbose: bool):
    """
    Checks all files necessary to push Generic Warehouse
    """
    storage = LocalStorage(directory=directory, with_timestamp=False)

    def _find(name: str) -> str:
        files = search_files(
            directory,
            filter_endswith=name,
            filter_extensions={"csv"},
        )
        assert files, f"Could not find files ending with {name}.csv"

        most_recent = max(files, key=os.path.getctime)
        _, filename, _ = explode(most_recent)
        return filename

    def _check(
        name: str,
        template: file_checker.FileTemplate,
        *,
        with_ids: bool = True,
    ) -> Ids:
        filename = _find(name)
        content = storage.get(filename)
        checker = file_checker.FileCheckerRun(
            content,
            template,
            filename,
            verbose,
        )
        checker.validate()
        if not with_ids:
            return set()
        return checker.occurrences(_ID_KEY)

    database_template = WarehouseTemplate.database()
    database_ids = _check("database", database_template)

    schema_template = WarehouseTemplate.schema(database_ids)
    schema_ids = _check("schema", schema_template)

    table_template = WarehouseTemplate.table(schema_ids)
    table_ids = _check("table", table_template)

    column_template = WarehouseTemplate.column(table_ids)
    _check("column", column_template, with_ids=False)

    user_template = WarehouseTemplate.user()
    user_ids = _check("user", user_template)

    view_ddl_template = WarehouseTemplate.view_ddl()
    _check("view_ddl", view_ddl_template, with_ids=False)

    query_template = WarehouseTemplate.query(database_ids, user_ids)
    _check("query", query_template, with_ids=False)


def main():
    parser = ArgumentParser()
    parser.add_argument(
        "-d",
        "--directory",
        help="Directory containing the files to be checked",
    )
    parser.add_argument(
        "--verbose",
        dest="display_issues",
        action="store_true",
        help="Show detailed logs",
    )
    args = parser.parse_args()
    process(args.directory, args.display_issues)



import logging
from datetime import datetime
from typing import Optional

from .types import TablesColumns
from .utils import build_path

logger = logging.getLogger(__name__)

EXCLUDED_DATABASES = {"system"}
EXCLUDED_SCHEMAS = {"information_schema", "default"}

TABLE_URL_TPL = "{host}explore/data/{catalog_name}/{schema_name}/{table_name}?o={workspace_id}"

TagMapping = dict[str, list[str]]


def _to_datetime_or_none(time_ms: Optional[int]) -> Optional[datetime]:
    """return time in ms as datetime or None"""
    if not time_ms:
        return None
    return datetime.fromtimestamp(time_ms / 1000.0)


def _table_payload(
    schema: dict,
    table: dict,
    host: str,
    workspace_id: str,
    tags: TagMapping,
) -> dict:
    """
    Prepares the table payload. This also includes a source link which is built
    here using the host and workspace_id.
    """
    url = TABLE_URL_TPL.format(
        host=host,
        catalog_name=table["catalog_name"],
        schema_name=table["schema_name"],
        table_name=table["name"],
        workspace_id=workspace_id,
    )

    keys = ["catalog_name", "schema_name", "name"]
    path = build_path(table, keys)

    return {
        "description": table.get("comment"),
        "id": table["table_id"],
        "owner_email": table.get("owner"),
        "schema_id": f"{schema['id']}",
        "table_name": table["name"],
        "tags": tags.get(path, []),
        "type": table.get("table_type"),
        "url": url,
    }


def _column_path(table: dict, column: dict) -> str:
    keys = ["catalog_name", "schema_name", "name"]
    table_path = build_path(table, keys)
    column_name = column["name"]
    return f"{table_path}.{column_name}"


def _column_payload(
    table: dict,
    column: dict,
    tags: TagMapping,
) -> dict:
    path = _column_path(table, column)
    return {
        "column_name": column["name"],
        "data_type": column["type_name"],
        "description": column.get("comment"),
        "id": f"`{table['table_id']}`.`{column['name']}`",
        "ordinal_position": column["position"],
        "table_id": table["table_id"],
        "tags": tags.get(path, []),
    }


class DatabricksFormatter:
    """
    Helper functions.
    Format the response in the format to be exported as csv
    """

    @staticmethod
    def format_database(raw_databases: list[dict]) -> list[dict]:
        databases = []
        for catalog in raw_databases:
            name = catalog["name"]
            if name in EXCLUDED_DATABASES:
                continue
            db = {
                "id": name,
                "database_name": name,
            }
            databases.append(db)
        return databases

    @staticmethod
    def format_schema(raw_schemas: list[dict], database: dict) -> list[dict]:
        schemas = []
        for schema in raw_schemas:
            if schema["name"] in EXCLUDED_SCHEMAS:
                continue
            sc = {
                "id": schema["full_name"],
                "database_id": database["id"],
                "schema_name": schema["name"],
                "description": schema.get("comment"),
                "tags": [],
            }
            schemas.append(sc)
        return schemas

    @staticmethod
    def format_table_column(
        raw_tables: list[dict],
        schema: dict,
        host: str,
        workspace_id: str,
        table_tags: TagMapping,
        column_tags: TagMapping,
    ) -> TablesColumns:
        tables = []
        columns = []
        if not raw_tables:
            return [], []
        for table in raw_tables:
            t = _table_payload(schema, table, host, workspace_id, table_tags)
            tables.append(t)
            if not table.get("columns"):
                continue
            for column in table["columns"]:
                c = _column_payload(table, column, column_tags)
                columns.append(c)

        return tables, columns

    @staticmethod
    def format_lineage(timestamps: dict) -> list[dict]:
        lineage: list[dict] = []
        for link, timestamp in timestamps.items():
            parent_path, child_path = link
            link_ = {
                "parent_path": parent_path,
                "child_path": child_path,
                "timestamp": timestamp,
            }
            lineage.append(link_)
        return lineage

    @staticmethod
    def format_query(raw_queries: list[dict]) -> list[dict]:
        queries = []
        for q in raw_queries:
            if not q["query_text"]:
                continue
            start_time = _to_datetime_or_none(q.get("query_start_time_ms"))
            end_time = _to_datetime_or_none(q.get("query_end_time_ms"))
            query = {
                "query_id": q["query_id"],
                "database_id": None,
                "database_name": None,
                "schema_name": None,
                "query_text": q["query_text"],
                "user_id": q["user_id"],
                "user_name": q.get("user_name"),
                "start_time": start_time,
                "end_time": end_time,
            }
            queries.append(query)
        return queries

    @staticmethod
    def _primary(emails: list[dict]) -> Optional[str]:
        """helper function to select a unique email"""
        if not emails:
            return None
        for email in emails:
            if email.get("primary", False):
                return email["value"]
        return emails[0]["value"]

    def _email(self, user: dict) -> Optional[str]:
        emails = user.get("emails")
        return self._primary(emails) if emails else None

    def format_user(self, raw_users: list[dict]) -> list[dict]:
        users = []
        for user in raw_users:
            users.append(
                {
                    "id": user["id"],
                    "email": self._email(user),
                    "first_name": None,
                    "last_name": user.get("displayName") or user["userName"],
                    "user_name": user["userName"],
                }
            )
        return users

    @staticmethod
    def format_view_ddl(tables: list[dict], schema: dict) -> list[dict]:
        view_ddl: list[dict] = []
        if not tables:
            return view_ddl
        for table in tables:
            if not table.get("view_definition"):
                continue
            view_ddl.append(
                {
                    "database_name": schema["database_id"],  # db id = name
                    "schema_name": schema["schema_name"],
                    "view_name": table["name"],
                    "view_definition": table["view_definition"],
                }
            )

        return view_ddl



from ...types import ExternalAsset, classproperty


class StrategyAsset(ExternalAsset):
    """Strategy assets that can be extracted"""

    ATTRIBUTE = "attribute"
    COLUMN = "column"
    CUBE = "cube"
    DASHBOARD = "dashboard"
    DOCUMENT = "document"
    FACT = "fact"
    LOGICAL_TABLE = "logical_table"
    METRIC = "metric"
    REPORT = "report"
    USER = "user"

    @classproperty
    def optional(cls) -> set["StrategyAsset"]:
        return {StrategyAsset.COLUMN}



"""
This file is a proxy to Looker SDK functions using Castor ApiSettings.
"""

from typing import Optional

from looker_sdk.rtl import transport
from looker_sdk.rtl.api_settings import ApiSettings, SettingsConfig
from looker_sdk.sdk.api40 import methods as methods40

from .credentials import LookerCredentials


def has_admin_permissions(sdk_: methods40.Looker40SDK) -> bool:
    """Check if any of the current user's permissions include admin rights"""
    current_user_id = sdk_.me().id
    if not current_user_id:
        return False
    current_user_roles = sdk_.user_roles(current_user_id)
    return any(
        "administer" in role.permission_set.permissions
        for role in current_user_roles
        if role.permission_set and role.permission_set.permissions
    )


class CastorApiSettings(ApiSettings):
    """SDK settings with initialisation using a credential object instead of a path to a .ini file"""

    def __init__(
        self, credentials: LookerCredentials, sdk_version: Optional[str] = ""
    ):
        """Configure using a config dict"""
        self.config = credentials.to_settings_config()
        self.verify_ssl = True
        self.base_url = self.config.get("base_url", "")
        self.timeout = credentials.timeout
        self.headers = {"Content-Type": "application/json"}
        self.agent_tag = f"{transport.AGENT_PREFIX}"
        if sdk_version:
            self.agent_tag += f" {sdk_version}"

        super().__init__()

    def read_config(self) -> SettingsConfig:
        """Returns a serialization of the credentials"""
        return self.config



import logging
from collections.abc import Iterable

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    write_json,
    write_summary,
)
from .assets import ModeAnalyticsAsset as Asset
from .client import Client, ModeCredentials

logger = logging.getLogger(__name__)


def iterate_all_data(
    client: Client,
) -> Iterable[tuple[Asset, list]]:
    """Iterate over the extracted Data From Mode Analytics"""

    datasources = client.fetch(Asset.DATASOURCE)
    yield Asset.DATASOURCE, deep_serialize(datasources)

    collections = client.fetch(Asset.COLLECTION)
    yield Asset.COLLECTION, deep_serialize(collections)

    reports = client.fetch(Asset.REPORT, additional_data=collections)
    yield Asset.REPORT, deep_serialize(reports)

    queries = client.fetch(Asset.QUERY, additional_data=reports)
    yield Asset.QUERY, deep_serialize(queries)

    members = client.fetch(Asset.MEMBER)
    yield Asset.MEMBER, deep_serialize(members)


def extract_all(**kwargs) -> None:
    """Extract Data From Mode Analytics and store it locally in files under the output_directory"""
    output_directory = kwargs.get("output") or from_env(OUTPUT_DIR)
    credentials = ModeCredentials(**kwargs)
    client = Client(credentials)

    ts = current_timestamp()

    for key, data in iterate_all_data(client):
        filename = get_output_filename(key.name.lower(), output_directory, ts)
        write_json(filename, data)

    write_summary(
        output_directory,
        ts,
        base_url=client.base_url(),
        client_name=client.name(),
    )



import logging
import random
import time
from collections.abc import Sequence
from enum import Enum
from typing import Any, Callable, Union

from pydantic import BaseModel, PositiveInt, PrivateAttr
from pydantic.fields import Field
from requests import HTTPError

logger = logging.getLogger(__name__)


MS_IN_SEC = 1000


class RetryStrategy(Enum):
    """retry mechanism"""

    CONSTANT = "CONSTANT"
    LINEAR = "LINEAR"
    EXPONENTIAL = "EXPONENTIAL"


DEFAULT_STRATEGY = RetryStrategy.CONSTANT


def _warning(callable: Callable, exception: BaseException) -> None:
    exception_type = type(exception)
    exception_path = f"{exception_type.__module__}.{exception_type.__name__}"
    callable_path = f"{callable.__module__}.{callable.__name__}"

    msg = f"Exception '{exception_path}' occurred within `{callable_path}`"
    logger.warning(msg)


class Retry(BaseModel):
    """
    This class checks if the retry conditions are met, and if so, how long to
    wait until the next attempt.
    """

    max_retries: PositiveInt
    base_ms: PositiveInt
    jitter_ms: PositiveInt
    strategy: RetryStrategy = Field(DEFAULT_STRATEGY, kw_only=True)
    _retry_attempts = PrivateAttr(default=0)

    def jitter(self) -> int:
        """compute random jitter in ms"""
        min_ = self.jitter_ms // 2
        max_ = int(self.jitter_ms * 1.5)
        return random.randrange(min_, max_)

    def base(self) -> int:
        """compute base wait time in ms"""
        base = self.base_ms
        if self.strategy == RetryStrategy.CONSTANT:
            return base
        if self.strategy == RetryStrategy.LINEAR:
            return base * self._retry_attempts
        # exponential
        scaled = float(base) / MS_IN_SEC
        return int(scaled**self._retry_attempts * MS_IN_SEC)

    def check(self, error: BaseException, log_exc_info: bool = False) -> bool:
        """
        Check whether retry should happen or not.
        If log_exc_info is True, it will add a more extensive log (most notably
        including the traceback).
        """
        if self._retry_attempts >= self.max_retries:
            return False

        exc_info = error if log_exc_info else None
        logger.warning("Caught a retryable exception", exc_info=exc_info)

        self._retry_attempts += 1
        wait_ms = self.base() + self.jitter()
        wait_s = float(wait_ms) / MS_IN_SEC
        msg = f"Attempting a new call in {wait_s} seconds, {self._retry_attempts} attempt(s) / {self.max_retries} max retries"
        logger.warning(msg)
        time.sleep(wait_s)
        return True


WrapperReturnType = Union[tuple[BaseException, None], tuple[None, Any]]


def retry(
    exceptions: Sequence[type[BaseException]],
    max_retries: int = 1,
    base_ms: int = 0,
    jitter_ms: int = 1,
    strategy: RetryStrategy = DEFAULT_STRATEGY,
    log_exc_info: bool = False,
) -> Callable:
    """retry decorator"""

    exceptions_ = tuple(e for e in exceptions)

    def _wrapper(callable: Callable) -> Callable:
        def _try(*args, **kwargs) -> WrapperReturnType:
            try:
                return None, callable(*args, **kwargs)
            except exceptions_ as err:
                _warning(callable, err)
                return err, None

        def _func(*args, **kwargs) -> Any:
            retry = Retry(
                max_retries=max_retries,
                base_ms=base_ms,
                jitter_ms=jitter_ms,
                strategy=strategy,
            )
            while True:
                err, result = _try(*args, **kwargs)
                if err is None:
                    return result
                if retry.check(err, log_exc_info):
                    continue
                raise err

        return _func

    return _wrapper


def retry_request(
    status_codes: Sequence[int],
    max_retries: int = 1,
    base_ms: int = 10,
    jitter_ms: int = 20,
    strategy: RetryStrategy = DEFAULT_STRATEGY,
    log_exc_info: bool = False,
) -> Callable:
    """retry decorator"""

    exceptions_ = tuple(e for e in status_codes)

    def _wrapper(callable: Callable) -> Callable:
        def _try(*args, **kwargs) -> WrapperReturnType:
            try:
                return None, callable(*args, **kwargs)
            except HTTPError as err:
                status_code = err.response.status_code
                if status_code not in exceptions_:
                    raise err
                _warning(callable, err)
                return err, None

        def _func(*args, **kwargs) -> Any:
            retry = Retry(
                max_retries=max_retries,
                base_ms=base_ms,
                jitter_ms=jitter_ms,
                strategy=strategy,
            )
            while True:
                err, result = _try(*args, **kwargs)
                if err is None:
                    return result
                if retry.check(err, log_exc_info):
                    continue
                raise err

        return _func

    return _wrapper



class NotionEndpointFactory:
    """Wrapper class around all endpoints we're using"""

    BLOCKS = "blocks"
    SEARCH = "search"
    USERS = "users"

    @classmethod
    def users(cls) -> str:
        return cls.USERS

    @classmethod
    def blocks(cls, block_id: str) -> str:
        return f"{cls.BLOCKS}/{block_id}/children"

    @classmethod
    def search(cls) -> str:
        return cls.SEARCH



from .abstract import AbstractSourceClient, SqlalchemyClient
from .api import (
    APIClient,
    Auth,
    BasicAuth,
    BearerAuth,
    CustomAuth,
    FetchNextPageBy,
    PaginationModel,
    RequestSafeMode,
    ResponseJson,
    build_url,
    fetch_all_pages,
    handle_response,
)
from .postgres import PostgresClient
from .query import ExtractionQuery
from .uri import uri_encode



import logging
from argparse import ArgumentParser

from castor_extractor.knowledge import notion  # type: ignore
from castor_extractor.utils import parse_filled_arguments  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    parser.add_argument("-t", "--token", help="Notion token")
    parser.add_argument("-o", "--output", help="Directory to write to")

    notion.extract_all(**parse_filled_arguments(parser))



from unittest.mock import MagicMock, patch

from .client import ConfluenceClient


def test_ConfluenceClient_folders():
    """
    Folder 1 -> Page A -> Folder 2 -> Folder 3 -> Folder 4 -> Page B
             -> Page C -> Folder 5 -> Page D

    After extracting the pages, we should have all IDs of folders that are
    immediate parents of pages. We still need to look out for nested folders.
    """
    folder_ids = {"1", "3", "4", "5"}
    mock_responses = {
        "1": {"id": "1", "parentType": None, "parentId": None},
        "2": {"id": "2", "parentType": "page", "parentId": "A"},
        "3": {"id": "3", "parentType": "folder", "parentId": "2"},
        "4": {"id": "4", "parentType": "folder", "parentId": "3"},
        "5": {"id": "5", "parentType": "page", "parentId": "C"},
    }

    def mock_get(endpoint):
        folder_id = endpoint.split("/")[-1]
        return mock_responses[folder_id]

    client = ConfluenceClient(credentials=MagicMock())

    with patch.object(client, "_get", side_effect=mock_get):
        result = list(client.folders(folder_ids))

    assert len(result) == 5
    assert {folder["id"] for folder in result} == set(mock_responses.keys())


def test_ConfluenceClient_filtered_spaces_with_allowlist():
    both_blocked_and_allowed_space_id = "789"
    archived_space_id = "934"
    random_space = "1000"

    spaces = [
        # Both blocked and allowed space. "Allowed" setting takes precedence.
        {
            "id": both_blocked_and_allowed_space_id,
            "type": "global",
            "status": "current",
        },
        {"id": archived_space_id, "type": "global", "status": "archived"},
        {"id": random_space, "type": "global", "status": "current"},
    ]
    # the "allowed" list overrides everything else
    client = ConfluenceClient(
        credentials=MagicMock(),
        include_archived_spaces=True,
        space_ids_allowed={
            both_blocked_and_allowed_space_id,
        },
        space_ids_blocked={
            both_blocked_and_allowed_space_id,
        },
    )

    with (
        patch(
            "source.packages.extractor.castor_extractor.knowledge.confluence.client.client.ConfluenceClient._get"
        ),
        patch(
            "source.packages.extractor.castor_extractor.knowledge.confluence.client.client.fetch_all_pages"
        ) as mock_fetch_all_pages,
    ):
        mock_fetch_all_pages.return_value = spaces

        filtered_spaces = list(client.spaces())

        assert len(filtered_spaces) == 1
        filtered_space_ids = {space["id"] for space in filtered_spaces}
        assert set(filtered_space_ids) == {both_blocked_and_allowed_space_id}


def test_ConfluenceClient_filtered_spaces():
    blocked_id = "42"
    personal_id = "666"
    archived_id = "934"
    random_id = "1000"

    # test the other settings : allow personal spaces & block space "42"
    client = ConfluenceClient(
        credentials=MagicMock(),
        include_archived_spaces=False,
        include_personal_spaces=True,
        space_ids_blocked={blocked_id},
    )

    spaces = [
        # Blocked space, to be skipped
        {"id": blocked_id, "type": "global", "status": "current"},
        # Archived space, to be skipped
        {"id": archived_id, "type": "collaboration", "status": "archived"},
        # Personal space, to be included
        {"id": personal_id, "type": "personal", "status": "current"},
        # Valid space
        {"id": random_id, "type": "knowledge_base", "status": "current"},
    ]

    with (
        patch(
            "source.packages.extractor.castor_extractor.knowledge.confluence.client.client.ConfluenceClient._get"
        ),
        patch(
            "source.packages.extractor.castor_extractor.knowledge.confluence.client.client.fetch_all_pages"
        ) as mock_fetch_all_pages,
    ):
        mock_fetch_all_pages.return_value = spaces

        filtered_spaces = list(client.spaces())

        filtered_space_ids = [space["id"] for space in filtered_spaces]

        # no duplicates
        assert len(filtered_space_ids) == len(set(filtered_space_ids))
        assert set(filtered_space_ids) == {personal_id, random_id}



from ...types import ExternalAsset


class ThoughtspotAsset(ExternalAsset):
    """Thoughtspot assets"""

    ANSWERS = "answers"
    ANSWER_USAGES = "answer_usages"
    LIVEBOARDS = "liveboards"
    LIVEBOARD_USAGES = "liveboard_usages"
    LOGICAL_TABLES = "logical_tables"



from typing import Optional


class CoalesceEndpointFactory:
    """Provide endpoints to hit Coalesce API"""

    @classmethod
    def environments(cls, environment_id: Optional[int] = None) -> str:
        """
        When specified, concatenate environment_id at the end to fetch details.
        Otherwise, list existing environments.
        """
        base = "api/v1/environments"
        if environment_id:
            return base + f"/{environment_id}"
        return base

    @classmethod
    def nodes(cls, environment_id: int, node_id: Optional[str] = None) -> str:
        """
        When specified, concatenate node_id at the end to fetch details.
        Otherwise, list existing nodes in the given environment.
        """
        base = f"api/v1/environments/{environment_id}/nodes"
        if node_id:
            return base + f"/{node_id}"
        return base

    @classmethod
    def runs(cls) -> str:
        """
        Get runs (additional filtering can be done in the body)
        """
        base = "api/v1/runs"
        return base

    @classmethod
    def run_results(cls, run_id: str) -> str:
        """
        get run results (including success/fail for tests), given a run id
        """
        return f"api/v1/runs/{run_id}/results"



import logging
from argparse import ArgumentParser

from castor_extractor.utils import parse_filled_arguments  # type: ignore
from castor_extractor.visualization import metabase  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    parser.add_argument("-b", "--base-url", help="Metabase base url")
    parser.add_argument("-u", "--user", help="Metabase username")
    parser.add_argument("-p", "--password", help="Metabase password")

    parser.add_argument("-o", "--output", help="Directory to write to")

    args = parse_filled_arguments(parser)

    credentials = metabase.MetabaseApiCredentials(**args)
    client = metabase.ApiClient(credentials)

    metabase.extract_all(client, **args)



from pydantic_settings import BaseSettings, SettingsConfigDict

CASTOR_ENV_PREFIX = "CASTOR_CONFLUENCE_"


class ConfluenceCredentials(BaseSettings):
    """Class to handle Confluence rest API permissions"""

    model_config = SettingsConfigDict(
        env_prefix=CASTOR_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    account_id: str
    base_url: str
    token: str
    username: str



from unittest.mock import patch

from requests import PreparedRequest, Request, Session

from .auth import BasicAuth
from .client import APIClient


class MockSession:
    def __init__(self):
        self.request_data = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def request(self, **kwargs) -> PreparedRequest:
        kwargs.pop("timeout")
        request = Request(**kwargs)
        prepared_request = Session().prepare_request(request=request)
        return prepared_request


@patch("requests.sessions.Session", MockSession)
def test__get():
    auth = BasicAuth(username="user_id", password="secret")
    client = APIClient(
        auth=auth,
        host="https://example.api.com/v1/",
        headers={"content-type": "test"},
        timeout=9,
    )

    prepared_request = client._call("GET", "endpoint")

    # test method
    assert prepared_request.method == "GET"

    # test headers
    assert prepared_request.headers["content-type"] == "test"
    assert (
        prepared_request.headers["Authorization"]
        == "Basic dXNlcl9pZDpzZWNyZXQ="
    )
    assert prepared_request.url == "https://example.api.com/v1/endpoint"



from ....utils import PaginationModel

_CLOUD_FIRST_PAGE = 0


class SodaCloudPagination(PaginationModel):
    content: list[dict]
    last: bool

    def is_last(self) -> bool:
        return self.last

    def next_page_payload(self) -> dict:
        current_page = (
            self.current_page_payload[self.fetch_by.value]["page"]
            if self.current_page_payload
            else _CLOUD_FIRST_PAGE
        )
        return {"page": current_page + 1}

    def page_results(self) -> list:
        return self.content



from typing import Any

from pydantic_settings import BaseSettings


class MetabaseLoginError(ConnectionError):
    def __init__(self, credentials_info: BaseSettings, error_details: Any):
        msg = "Connection to Metabase failed!\n"
        msg += f"Credentials: {credentials_info}\n"
        msg += f"Details: {error_details}\n"
        super().__init__(msg)


class SuperuserCredentialsRequired(AssertionError):
    def __init__(self, credentials_info: BaseSettings, error_details: Any):
        msg = "Superuser credentials are required!\n"
        msg += f"Credentials: {credentials_info}\n"
        msg += f"Details: {error_details}\n"
        super().__init__(msg)


class EncryptionSecretKeyRequired(AssertionError):
    """
    Raised when missing the encryption secret key while it was required.
    """

    def __init__(self, credentials_info: BaseSettings, error_details: Any):
        msg = "Encryption secret key is required!\n"
        msg += f"Credentials: {credentials_info}\n"
        msg += f"Details: {error_details}\n"
        super().__init__(msg)



from typing import Optional

from pydantic import ConfigDict, Field
from pydantic.alias_generators import to_camel

from ....utils import PaginationModel

SIGMA_API_LIMIT = 1_000  # default number of records per page
SIGMA_QUERIES_PAGINATION_LIMIT = 50


class SigmaPagination(PaginationModel):
    next_page: Optional[str] = "0"
    entries: list = Field(default_factory=list)

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    def is_last(self) -> bool:
        return self.next_page is None

    def next_page_payload(self) -> dict:
        return {"page": self.next_page}

    def page_results(self) -> list:
        return self.entries



import logging
from argparse import ArgumentParser

from castor_extractor.utils import parse_filled_arguments  # type: ignore
from castor_extractor.visualization import salesforce_reporting  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    parser.add_argument("-u", "--username", help="Salesforce username")
    parser.add_argument("-p", "--password", help="Salesforce password")
    parser.add_argument("-c", "--client-id", help="Salesforce client id")
    parser.add_argument(
        "-s", "--client-secret", help="Salesforce client secret"
    )
    parser.add_argument(
        "-t", "--security-token", help="Salesforce security token"
    )
    parser.add_argument("-b", "--base-url", help="Salesforce instance URL")
    parser.add_argument("-o", "--output", help="Directory to write to")

    salesforce_reporting.extract_all(**parse_filled_arguments(parser))



import logging
from abc import abstractmethod
from collections.abc import Iterator
from enum import Enum
from functools import partial
from time import sleep
from typing import Callable, Optional, Union

from pydantic import BaseModel

logger = logging.getLogger(__name__)


class FetchNextPageBy(Enum):
    """
    Enum to pick which APIClient._call() argument we want
    to use for calling the next page in the pagination.
    Supported arguments are :
    - params (PAYLOAD)
    - endpoint (URL)
    """

    PAYLOAD = "pagination_params"
    URL = "endpoint"


class PaginationModel(BaseModel):
    """
    Base abstract class defining a pagination model

    By implementing the 3 abstract methods below, enables
    to fetch all elements of a Paginated API by using the
    `fetch_all_pages` method
    """

    fetch_by: FetchNextPageBy = FetchNextPageBy.PAYLOAD
    current_page_payload: Optional["dict"] = None

    @abstractmethod
    def is_last(self) -> bool:
        """Stopping condition for the pagination"""
        pass

    @abstractmethod
    def next_page_payload(self) -> Optional[Union[dict, str]]:
        """Payload enabling to generate the request for the next page"""
        pass

    @abstractmethod
    def page_results(self) -> list:
        """List of results of the current page"""
        pass

    def next_page_parameters(self) -> dict:
        return {self.fetch_by.value: self.next_page_payload()}


def fetch_all_pages(
    request: Callable,
    pagination_model: type[PaginationModel],
    rate_limit: Optional[int] = None,
) -> Iterator:
    """
    Method to return all results of a Paginated API based on the
    pagination model and the first request call
    """
    page_number = 1
    response_payload = request()

    paginated_response = pagination_model(**response_payload)

    while not paginated_response.is_last():
        logger.debug(f"Fetching page number {page_number}")
        yield from paginated_response.page_results()
        next_page_parameters = paginated_response.next_page_parameters()
        request_with_pagination = partial(request, **next_page_parameters)
        if rate_limit:
            sleep(rate_limit)
        paginated_response = pagination_model(
            current_page_payload=next_page_parameters,
            **request_with_pagination(),
        )
        page_number += 1

    # send last page's results
    logger.debug(f"Fetching page number {page_number}")
    yield from paginated_response.page_results()



import logging
from collections.abc import Iterable
from typing import Union

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    write_json,
    write_summary,
)
from ...utils.salesforce import SalesforceCredentials
from .assets import SalesforceReportingAsset
from .client import SalesforceReportingClient

logger = logging.getLogger(__name__)


def iterate_all_data(
    client: SalesforceReportingClient,
) -> Iterable[tuple[str, Union[list, dict]]]:
    """Iterate over the extracted data from Salesforce"""

    for asset in SalesforceReportingAsset:
        logger.info(f"Extracting {asset.value.upper()} from REST API")
        data = client.fetch(asset)
        yield asset.name.lower(), deep_serialize(data)


def extract_all(**kwargs) -> None:
    """
    Extract data from Salesforce REST API
    Store the output files locally under the given output_directory
    """
    _output_directory = kwargs.get("output") or from_env(OUTPUT_DIR)
    credentials = SalesforceCredentials(**kwargs)

    client = SalesforceReportingClient(credentials=credentials)
    ts = current_timestamp()

    for key, data in iterate_all_data(client):
        filename = get_output_filename(key, _output_directory, ts)
        write_json(filename, data)

    write_summary(_output_directory, ts)



from pydantic import Field, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from ....utils import OUTPUT_DIR
from ..constants import (
    DEFAULT_LOOKER_PAGE_SIZE,
    DEFAULT_LOOKER_THREAD_POOL_SIZE,
    LOOKER_ENV_PREFIX,
    MAX_THREAD_POOL_SIZE,
    MIN_THREAD_POOL_SIZE,
)


class ExtractionParameters(BaseSettings):
    """
    Class holding all the parameters needed for the extraction of
    Looker metadata
    """

    model_config = SettingsConfigDict(
        env_prefix=LOOKER_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    is_safe_mode: bool = False
    log_to_stdout: bool
    output: str = Field(validation_alias=OUTPUT_DIR)
    search_per_folder: bool
    page_size: int = Field(default=DEFAULT_LOOKER_PAGE_SIZE)
    thread_pool_size: int = Field(default=DEFAULT_LOOKER_THREAD_POOL_SIZE)

    @field_validator("thread_pool_size", mode="before")
    @classmethod
    def _check_thread_pool_size(cls, thread_pool_size: int) -> int:
        thread_pool_size = thread_pool_size or DEFAULT_LOOKER_THREAD_POOL_SIZE
        if MIN_THREAD_POOL_SIZE <= thread_pool_size <= MAX_THREAD_POOL_SIZE:
            return thread_pool_size
        raise ValueError("Thread pool size must be between 1 and 200 inclusive")



import logging
from collections.abc import Iterator
from concurrent.futures import ThreadPoolExecutor
from typing import Optional

import requests

from ....utils import build_url
from ....utils.salesforce import SalesforceBaseClient
from ..assets import SalesforceReportingAsset
from .soql import queries

logger = logging.getLogger(__name__)

REQUIRING_URL_ASSETS = (
    SalesforceReportingAsset.REPORTS,
    SalesforceReportingAsset.DASHBOARDS,
    SalesforceReportingAsset.FOLDERS,
)

_CONCURRENT_THREADS = 50


class SalesforceReportingClient(SalesforceBaseClient):
    """
    Salesforce Reporting API client
    """

    def _get_asset_url(
        self, asset_type: SalesforceReportingAsset, asset: dict
    ) -> Optional[str]:
        """
        Fetch the given Asset + add the corresponding URL.
        """

        if asset_type == SalesforceReportingAsset.DASHBOARDS:
            path = f"lightning/r/Dashboard/{asset['Id']}/view"
            return build_url(self._host, path)

        if asset_type == SalesforceReportingAsset.FOLDERS:
            path = asset["attributes"]["url"].lstrip("/")
            return build_url(self._host, path)

        if asset_type == SalesforceReportingAsset.REPORTS:
            path = f"lightning/r/Report/{asset['Id']}/view"
            return build_url(self._host, path)

        return None

    def _fetch_and_add_url(
        self, asset_type: SalesforceReportingAsset
    ) -> Iterator[dict]:
        assets = self._query_all(queries[asset_type])
        for asset in assets:
            url = self._get_asset_url(asset_type, asset)
            yield {**asset, "Url": url}

    def _metadata(self, report_id: str) -> Optional[dict]:
        url = f"services/data/v60.0/analytics/reports/{report_id}/describe"
        try:
            metadata = self._get(url, retry_on_timeout=False)
            # pick only what we need to build the lineage
            columns = metadata["reportExtendedMetadata"]["detailColumnInfo"]
            return {
                "reportId": report_id,
                "detailColumnInfo": columns or dict(),
            }
        except (requests.HTTPError, requests.RequestException) as ex:
            # Extracting column metadata is used only for lineage purposes
            # and is non-critical. API errors are common during this step,
            # so we choose to skip them rather than fail the process. The same
            # rows consistently fail, and retries have proven ineffective.
            logger.info(ex)
            return None

    def _fetch_reports_metadata(self) -> Iterator[dict]:
        """
        Use the "describe" endpoint to extract report metadata.
        Keep only the detailColumnInfo, which is required for building the lineage.

        More info here:
        https://developer.salesforce.com/docs/atlas.en-us.api_analytics.meta/api_analytics/sforce_analytics_rest_api_getbasic_reportmetadata.htm
        https://www.notion.so/castordoc/Salesforce-Lineage-216a1c3d458580859888cf4ca2d7fa51?source=copy_link
        """
        # The "describe" endpoint requires report_ids. To avoid introducing
        # task dependencies, we opted to re-extract the reports.
        # It is fast anyway, since it's running a SQL query
        reports = self.fetch(SalesforceReportingAsset.REPORTS)
        report_ids = [report["Id"] for report in reports]

        # Calling "describe" on each report individually can be slow,
        # especially for accounts with thousands of reports. That's why
        # we use multithreading here — it significantly improves performance.
        with ThreadPoolExecutor(max_workers=_CONCURRENT_THREADS) as executor:
            fetch_results = executor.map(self._metadata, report_ids)

            for metadata in fetch_results:
                if not metadata:
                    continue
                yield metadata

    def fetch(self, asset: SalesforceReportingAsset) -> list[dict]:
        """
        Fetch Salesforce Reporting assets
        """
        logger.info(f"Starting extraction of {asset}")

        if asset in REQUIRING_URL_ASSETS:
            return list(self._fetch_and_add_url(asset))

        if asset == SalesforceReportingAsset.REPORTS_METADATA:
            return list(self._fetch_reports_metadata())

        return list(self._query_all(queries[asset]))



from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

SODA_ENV_PREFIX = "CASTOR_SODA_"

_CLOUD_API = "https://cloud.soda.io/api/v1/"


class SodaCredentials(BaseSettings):
    """Class to handle Soda rest API permissions"""

    model_config = SettingsConfigDict(
        env_prefix=SODA_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    api_key: str = Field(repr=False)
    secret: str = Field(repr=False)
    host: str = Field(default=_CLOUD_API)



from collections.abc import Iterable, Iterator
from typing import Any

import tableauserverclient as TSC  # type: ignore

from ....utils import JsonType, SerializedAsset
from ..assets import TableauAsset
from .rest_fields import REST_FIELDS


def _pick(element: Any, key: str) -> JsonType:
    if isinstance(element, dict):
        return element[key]
    else:
        return getattr(element, key)


class TableauClientTSC:
    """
    Extract Tableau Assets using TableauServerClient (TSC)
    https://tableau.github.io/server-client-python/docs/api-ref
    """

    def __init__(
        self,
        server: TSC.Server,
    ):
        self._server = server

    def _pick_fields(
        self,
        data: Iterable,
        asset: TableauAsset,
    ) -> Iterator[dict]:
        keys = REST_FIELDS[asset]

        for row in data:
            fields = {key: _pick(row, key) for key in keys}
            if asset == TableauAsset.USER:
                self._server.users.populate_groups(row)
                fields["group_ids"] = [group.id for group in row.groups]

            yield fields

    def fetch(
        self,
        asset: TableauAsset,
    ) -> SerializedAsset:
        if asset == TableauAsset.DATASOURCE:
            data = TSC.Pager(self._server.datasources)

        elif asset == TableauAsset.PROJECT:
            data = TSC.Pager(self._server.projects)

        elif asset == TableauAsset.USAGE:
            data = TSC.Pager(self._server.views, usage=True)

        elif asset == TableauAsset.USER:
            data = TSC.Pager(self._server.users)

        elif asset == TableauAsset.WORKBOOK:
            data = TSC.Pager(self._server.workbooks)

        else:
            raise AssertionError(f"Fetching from TSC not supported for {asset}")

        return list(self._pick_fields(data, asset))



import logging
from http import HTTPStatus
from typing import Callable, Literal, Optional

import requests
from requests import Response

from ...retry import retry, retry_request
from .auth import Auth
from .safe_request import RequestSafeMode, handle_response
from .utils import build_url

logger = logging.getLogger(__name__)

Headers = Optional[dict[str, str]]

# https://requests.readthedocs.io/en/latest/api/#requests.request
HttpMethod = Literal["GET", "OPTIONS", "HEAD", "POST", "PUT", "PATCH", "DELETE"]

DEFAULT_TIMEOUT = 60
RETRY_ON_EXPIRED_TOKEN = 1
RETRY_ON_GATEWAY_TIMEOUT = 3

_TIMEOUT_RETRY_BASE_MS = 10 * 60 * 1000  # 10 minutes
_TIMEOUT_RETRY_COUNT = 2
_TIMEOUT_RETRY_EXCEPTIONS = (requests.exceptions.Timeout,)


def _generate_payloads(
    method: HttpMethod,
    params: Optional[dict],
    data: Optional[dict],
    pagination_params: Optional[dict],
) -> tuple[Optional[dict], Optional[dict]]:
    _pagination_params = pagination_params or {}

    if method == "GET":
        params = params or {}
        params = {**params, **_pagination_params}
        return data, params
    if method == "POST":
        data = data or {}
        data = {**data, **_pagination_params}
        return data, params
    raise ValueError(f"Method {method} is not yet supported")


class APIClient:
    """
    Interface to easily query REST-API with GET and POST requests

    Args:
        auth: auth class to enable logging to the API
        host: base url of the API
        headers: common headers to all calls that will be made
        timeout: read timeout for each request
        safe_mode: ignore certain exceptions based on status codes

    Note:
        If the auth implements a refreshing mechanism (refresh_token)
        the token is automatically refreshed once upon receiving the
        401: UNAUTHORIZED status code
    """

    def __init__(
        self,
        auth: Auth,
        host: Optional[str] = None,
        headers: Headers = None,
        timeout: int = DEFAULT_TIMEOUT,
        safe_mode: RequestSafeMode = RequestSafeMode(),
    ):
        self.base_headers = headers or {}
        self._host = host
        self._timeout = timeout
        self._auth = auth
        self._safe_mode = safe_mode

    def _call(
        self,
        method: HttpMethod,
        endpoint: str,
        *,
        headers: Headers = None,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        pagination_params: Optional[dict] = None,
        retry_on_timeout: bool = True,
    ) -> Response:
        headers = headers or {}

        data, params = _generate_payloads(
            method=method,
            params=params,
            data=data,
            pagination_params=pagination_params,
        )

        url = build_url(self._host, endpoint)

        if retry_on_timeout:
            retry_wrapper = retry(
                exceptions=_TIMEOUT_RETRY_EXCEPTIONS,
                max_retries=_TIMEOUT_RETRY_COUNT,
                base_ms=_TIMEOUT_RETRY_BASE_MS,
            )
            request_fn = retry_wrapper(requests.request)
        else:
            request_fn = requests.request

        return request_fn(
            method=method,
            url=url,
            auth=self._auth,
            headers={**self.base_headers, **headers},
            params=params,
            json=data,
            timeout=self._timeout,
        )

    @retry_request(
        status_codes=(HTTPStatus.GATEWAY_TIMEOUT,),
        max_retries=RETRY_ON_GATEWAY_TIMEOUT,
    )
    @retry_request(
        status_codes=(HTTPStatus.UNAUTHORIZED,),
        max_retries=RETRY_ON_EXPIRED_TOKEN,
    )
    def _get(
        self,
        endpoint: str,
        *,
        headers: Headers = None,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        pagination_params: Optional[dict] = None,
        retry_on_timeout: bool = True,
    ):
        response = self._call(
            method="GET",
            endpoint=endpoint,
            params=params,
            data=data,
            pagination_params=pagination_params,
            headers=headers,
            retry_on_timeout=retry_on_timeout,
        )
        if response.status_code == HTTPStatus.UNAUTHORIZED:
            self._auth.refresh_token()

        return handle_response(response, safe_mode=self._safe_mode)

    @retry_request(
        status_codes=(HTTPStatus.UNAUTHORIZED,),
        max_retries=RETRY_ON_EXPIRED_TOKEN,
    )
    def _post(
        self,
        endpoint: str,
        *,
        headers: Headers = None,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        pagination_params: Optional[dict] = None,
        handler: Optional[Callable] = None,
    ):
        response = self._call(
            method="POST",
            endpoint=endpoint,
            params=params,
            data=data,
            pagination_params=pagination_params,
            headers=headers,
        )
        if response.status_code == HTTPStatus.UNAUTHORIZED:
            self._auth.refresh_token()

        return handle_response(
            response, safe_mode=self._safe_mode, handler=handler
        )



import json
import logging
from collections.abc import Iterable
from typing import Optional, Union, cast

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    write_json,
    write_summary,
)
from .assets import LookerStudioAsset
from .client import LookerStudioClient, LookerStudioCredentials

logger = logging.getLogger(__name__)

APPLICATION_CREDENTIALS = "GOOGLE_APPLICATION_CREDENTIALS"
LOOKER_STUDIO_ADMIN_EMAIL = "CASTOR_LOOKER_STUDIO_ADMIN_EMAIL"


def iterate_all_data(
    client: LookerStudioClient,
    has_view_activity_logs: bool = True,
) -> Iterable[tuple[LookerStudioAsset, Union[list, dict]]]:
    assets_to_extract = LookerStudioAsset.mandatory

    if has_view_activity_logs:
        assets_to_extract.add(LookerStudioAsset.VIEW_ACTIVITY)

    for asset in assets_to_extract:
        logger.info(f"Extracting {asset.name} from API")
        data = list(deep_serialize(client.fetch(asset)))
        yield asset, data
        logger.info(f"Extracted {len(data)} {asset.name} from API")


def _credentials(params: dict) -> LookerStudioCredentials:
    """
    Builds the Looker Studio credentials by combining the Service Account
    credentials with the admin email.
    """
    path = params.get("credentials") or from_env(APPLICATION_CREDENTIALS)
    logger.info(f"Looker Studio credentials loaded from {path}")
    with open(path) as file:
        credentials = cast(dict, json.load(file))

    admin_email = params.get("admin_email") or from_env(
        LOOKER_STUDIO_ADMIN_EMAIL
    )
    credentials["admin_email"] = admin_email
    has_view_activity_logs = not params["skip_view_activity_logs"]
    credentials["has_view_activity_logs"] = has_view_activity_logs
    return LookerStudioCredentials(**credentials)


def _bigquery_credentials_or_none(params: dict) -> Optional[dict]:
    """Extracts optional GCP credentials to access BigQuery"""
    path = params.get("bigquery_credentials") or from_env(
        APPLICATION_CREDENTIALS,
        allow_missing=True,
    )
    if not path:
        return None

    logger.info(f"BigQuery credentials loaded from {path}")
    with open(path) as file:
        return cast(dict, json.load(file))


def _validate_user_emails(user_emails: list[str]):
    """
    Raises an error if the user emails are not in the expected format (list of strings),
    or if the list is empty.
    """
    if not isinstance(user_emails, list):
        raise TypeError("The users file must be a list")

    if len(user_emails) == 0:
        raise ValueError("The users file must contain at least one user email")

    if not all(isinstance(email, str) for email in user_emails):
        raise TypeError("All items in users list must be strings")


def _read_optional_user_emails(
    users_file_path: Optional[str],
) -> Optional[list[str]]:
    """Loads the user emails from a file, if it was provided."""
    if not users_file_path:
        return None

    with open(users_file_path, "r") as file:
        user_emails = json.load(file)

    _validate_user_emails(user_emails)
    return user_emails


def extract_all(**kwargs) -> None:
    """
    Extracts data from Looker Studio and stores the output files locally under
    the given output_directory.
    """
    users_file_path = kwargs.get("users_file_path")
    output_directory = kwargs.get("output") or from_env(OUTPUT_DIR)

    credentials = _credentials(kwargs)
    has_view_activity_logs = bool(credentials.has_view_activity_logs)
    user_emails = _read_optional_user_emails(users_file_path)

    bigquery_credentials = _bigquery_credentials_or_none(kwargs)

    client = LookerStudioClient(
        credentials=credentials,
        bigquery_credentials=bigquery_credentials,
        user_emails=user_emails,
    )
    ts = current_timestamp()

    for key, data in iterate_all_data(client, has_view_activity_logs):
        filename = get_output_filename(key.name.lower(), output_directory, ts)
        write_json(filename, data)

    write_summary(output_directory, ts)



class Keys:
    ACCESS_TOKEN = "access_token"  # noqa: S105
    ACTIVITY_EVENT_ENTITIES = "activityEventEntities"
    CONTINUATION_URI = "continuationUri"
    ID = "id"
    INACTIVE_WORKSPACES = "excludeInActiveWorkspaces"
    LAST_RESULT_SET = "lastResultSet"
    PERSONAL_WORKSPACES = "excludePersonalWorkspaces"
    STATUS = "status"
    VALUE = "value"
    WORKSPACES = "workspaces"



from collections import defaultdict
from collections.abc import Iterable, Sequence
from typing import (
    Any,
    List,
    Optional,
    TypeVar,
)

from .object import getproperty
from .type import Getter

T = TypeVar("T")


def group_by(identifier: Getter, elements: Sequence) -> dict[Any, list]:
    """Groups the elements by the given key"""
    groups: dict[Any, list] = defaultdict(list)
    for element in elements:
        key = getproperty(element, identifier)
        groups[key].append(element)

    return groups


def mapping_from_rows(rows: list[dict], key: Any, value: Any) -> dict:
    """
    Create a dictionary mapping from a list of dictionaries using specified keys for mapping.

    Args:
        rows (list[dict]): A list of dictionaries from which to create the mapping.
        key (Any): The key to use for the keys of the resulting dictionary.
        value (Any): The key to use for the values of the resulting dictionary.

    Returns:
        dict: A dictionary where each key-value pair corresponds to the specified key and value
              from each dictionary in the input list. Only dictionaries with both specified key
              and value present are included in the result.

    Example:
        rows = [{'id': 1, 'name': 'Alice'}, {'id': 2, 'name': 'Bob'}]
        mapping = mapping_from_rows(rows, 'id', 'name')
        # mapping will be {1: 'Alice', 2: 'Bob'}
    """
    mapping = {}

    for row in rows:
        mapping_key = row.get(key)
        mapping_value = row.get(value)

        if not mapping_key or not mapping_value:
            continue
        mapping[mapping_key] = mapping_value

    return mapping


def empty_iterator():
    """
    Utils to return empty iterator, mainly used for viz transformers
    Remark: missing return type is on purpose, it breaks the typing
    """
    return iter([])


def deduplicate(
    identifier: Getter,
    elements: Iterable[T],
) -> list[T]:
    """
    Remove duplicates in the given elements, using the specified identifier
    Only the first occurrence is kept.
    """
    deduplicated: list[T] = []
    processed: set[Any] = set()

    for element in elements:
        key = getproperty(element, identifier)
        if key in processed:
            continue
        processed.add(key)
        deduplicated.append(element)

    return deduplicated


def filter_items(
    items: Iterable[T],
    allowed: Optional[Iterable[T]] = None,
    blocked: Optional[Iterable[T]] = None,
) -> List[T]:
    """
    Filters `items` by excluding any in `blocked` or including only those in `allowed`.
    If both `allowed` and `blocked` are None, returns all items.
    If both are provided, raise an error.
    """
    items = list(items)

    if allowed and blocked:
        raise AttributeError(
            "Only one of `allowed` and `blocked` can be provided"
        )
    if blocked:
        return [item for item in items if item not in blocked]
    if allowed:
        return [item for item in items if item in allowed]

    return items



class DatabricksEndpointFactory:
    @classmethod
    def tables(cls):
        return "api/2.1/unity-catalog/tables"

    @classmethod
    def schemas(cls):
        return "api/2.1/unity-catalog/schemas"

    @classmethod
    def databases(cls):
        return "api/2.1/unity-catalog/catalogs"

    @classmethod
    def table_lineage(cls):
        return "api/2.0/lineage-tracking/table-lineage"

    @classmethod
    def column_lineage(cls):
        return "api/2.0/lineage-tracking/column-lineage"

    @classmethod
    def queries(cls):
        return "api/2.0/sql/history/queries"

    @classmethod
    def users(cls):
        return "api/2.0/preview/scim/v2/Users"



import os
from abc import ABC, abstractmethod
from collections.abc import Iterable, Iterator
from io import StringIO
from typing import Optional

from .constants import ENCODING_UTF8
from .formatter import CsvFormatter, Formatter
from .time import current_timestamp
from .write import timestamped_filename


class AbstractStorage(ABC):
    """Abstract class to store warehouse assets"""

    @abstractmethod
    def path(self, name: str) -> str:
        """builds the path associated to the given name (it doesn't necessarily exist)"""

    @abstractmethod
    def exists(self, name: str) -> bool:
        """checks file existence"""

    @abstractmethod
    def put(self, name: str, data: Iterable[dict]) -> str:
        """put the given data in the storage and returns its path"""

    @abstractmethod
    def get(self, name: str) -> Iterable[dict]:
        """read data"""


class LocalStorage(AbstractStorage):
    """Store assets in local file system"""

    def __init__(
        self,
        directory: str,
        formatter: Optional[Formatter] = None,
        with_timestamp: bool = True,
    ):
        self._directory = directory
        self._formatter = CsvFormatter() if formatter is None else formatter
        self.stored_at_ts = current_timestamp()
        self._with_timestamp = with_timestamp

    def path(self, name: str) -> str:
        filename = f"{name}.{self._formatter.extension()}"
        if self._with_timestamp:
            filename = timestamped_filename(filename, self.stored_at_ts)
        return os.path.join(
            self._directory,
            filename,
        )

    def exists(self, name: str) -> bool:
        path = self.path(name)
        return os.path.isfile(path)

    def put(self, name: str, data: Iterable[dict]) -> str:
        path = self.path(name)
        with open(path, "w", encoding=ENCODING_UTF8) as file:
            self._formatter.serialize(file, data)
        return path

    def get(self, name: str) -> Iterator[dict]:
        path = self.path(name)
        with open(path, "r", encoding=ENCODING_UTF8) as file:
            return self._formatter.deserialize(StringIO(file.read()))



import logging
from argparse import ArgumentParser

from castor_extractor.uploader import (  # type: ignore
    FileType,
    Zone,
    upload_any,
)
from castor_extractor.utils import parse_filled_arguments  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def _args() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        "-k",
        "--token",
        help="API token provided by CastorDoc",
    )
    parser.add_argument(
        "-s",
        "--source_id",
        help="source id provided by CastorDoc",
    )
    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument("-f", "--file_path", help="path to file to upload")
    group.add_argument(
        "-d",
        "--directory_path",
        help="""directory containing the files to upload.
                WARNING: it will upload all the files included
                in the given repository.""",
    )
    supported_file_type = [ft.value for ft in FileType]
    parser.add_argument(
        "-t",
        "--file_type",
        help="type of file to upload, currently supported are {}".format(
            supported_file_type,
        ),
        choices=supported_file_type,
    )
    supported_zones = [zone.value for zone in Zone]
    parser.add_argument(
        "-z",
        "--zone",
        help="geographic zone to upload, currently supported are {}, defaults to EU".format(
            supported_zones,
        ),
        choices=supported_zones,
    )
    return parser


def main():
    parser = _args()
    upload_any(**parse_filled_arguments(parser))



import logging
from argparse import ArgumentParser

from castor_extractor.utils import parse_filled_arguments  # type: ignore
from castor_extractor.visualization import powerbi  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    parser.add_argument("-t", "--tenant_id", help="PowerBi tenant ID")
    parser.add_argument("-c", "--client_id", help="PowerBi client ID")
    parser.add_argument("-s", "--secret", help="PowerBi password")
    parser.add_argument(
        "-sc",
        "--scopes",
        help="PowerBi scopes, optional",
        nargs="*",
    )
    parser.add_argument("-o", "--output", help="Directory to write to")
    parser.add_argument("-l", "--login_url", help="Login url (Optional)")
    parser.add_argument("-a", "--api_base", help="REST API base (Optional)")

    powerbi.extract_all(**parse_filled_arguments(parser))



from dataclasses import dataclass
from typing import Optional

PER_PAGE = 50  # maximum value accepted by DOMO is 50


@dataclass
class Pagination:
    """Handles pagination within DOMO Api"""

    number_results: Optional[int] = None
    offset: int = 0
    per_page: int = PER_PAGE
    should_stop: bool = False

    @property
    def needs_increment(self) -> bool:
        if self.number_results is None:
            return True  # first iteration

        if (self.number_results < self.per_page) or self.should_stop:
            return False

        return True

    @property
    def params(self) -> dict:
        return {"offset": self.offset, "limit": self.per_page}

    def increment_offset(self, number_results: int) -> None:
        self.offset += number_results
        self.number_results = number_results



from typing import Optional

from pydantic import ConfigDict, Field
from pydantic.alias_generators import to_camel

from .pagination import PaginationModel, fetch_all_pages


class _TestPagination(PaginationModel):
    next_page: Optional[str] = None
    entries: list = Field(default_factory=list)

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    def is_last(self) -> bool:
        return self.next_page is None

    def next_page_payload(self) -> dict:
        return {"page": self.next_page}

    def page_results(self) -> list:
        return self.entries


def _request(pagination_params: Optional[dict] = None):
    if not pagination_params:
        return {
            "nextPage": "next_page_id",
            "entries": [1, 2, 3, 4, 5],
        }
    if pagination_params.get("page") == "next_page_id":
        return {
            "nextPage": "next_page_id_2",
            "entries": [6, 7, 8, 9, 10],
        }
    if pagination_params.get("page") == "next_page_id_2":
        return {
            "nextPage": None,
            "entries": [11],
        }

    raise AssertionError(f"call has unexpected parameters: {pagination_params}")


def test__TestPagination():
    all_results = fetch_all_pages(_request, _TestPagination)
    assert list(all_results) == [i for i in range(1, 12)]



from typing import Optional

from pydantic import Field

from ...utils import PaginationModel

DATABRICKS_PAGE_SIZE = 100


class DatabricksPagination(PaginationModel):
    next_page_token: Optional[str] = None
    has_next_page: bool = False
    res: list[dict] = Field(default_factory=list)

    def is_last(self) -> bool:
        return not (self.has_next_page and self.next_page_token)

    def next_page_payload(self) -> dict:
        return {"page_token": self.next_page_token}

    def page_results(self) -> list:
        return self.res



class ConfluenceEndpointFactory:
    """
    Confluence rest api v2 endpoint factory.
    https://developer.atlassian.com/cloud/confluence/rest/v2/intro/#about
    """

    API = "wiki/api/v2/"
    DATABASE = "databases"
    FOLDERS = "folders"
    PAGES = "pages"
    SPACES = "spaces"
    USERS = "users-bulk"

    @classmethod
    def database(cls, database_id: str) -> str:
        """
        Endpoint to fetch a database by id.
        More: https://developer.atlassian.com/cloud/confluence/rest/v2/api-group-database/#api-databases-id-get
        """
        return f"{cls.API}{cls.DATABASE}/{database_id}"

    @classmethod
    def folder(cls, folder_id: str) -> str:
        """
        Endpoint to fetch a folder by id.
        More: https://developer.atlassian.com/cloud/confluence/rest/v2/api-group-folder/#api-folders-id-get
        """
        return f"{cls.API}{cls.FOLDERS}/{folder_id}"

    @classmethod
    def pages(cls, space_id: str) -> str:
        """
        Endpoint to fetch all pages in the given space.
        More: https://developer.atlassian.com/cloud/confluence/rest/v2/api-group-page/#api-spaces-id-pages-get
        """
        return f"{cls.API}{cls.SPACES}/{space_id}/{cls.PAGES}?body-format=atlas_doc_format"

    @classmethod
    def spaces(cls) -> str:
        """
        Endpoint to fetch all spaces.
        https://developer.atlassian.com/cloud/confluence/rest/v2/api-group-space/#api-spaces-get
        """
        return f"{cls.API}{cls.SPACES}"

    @classmethod
    def users(cls) -> str:
        """
        Endpoint to fetch all user.
        More: https://developer.atlassian.com/cloud/confluence/rest/v2/api-group-user/#api-users-bulk-post
        """
        return f"{cls.API}{cls.USERS}"



from typing import Optional

from pydantic import field_validator, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

# To specify the default site on Tableau Server, you can use an empty string
# https://tableau.github.io/server-client-python/docs/api-ref#authentication
_DEFAULT_SERVER_SITE_ID = ""

# In Castor APP, site_id is mandatory: users can't let this field empty
# In that case, we encourage users to write "Default" instead
_DEFAULT_SITE_ID_USER_INPUT = "default"


TABLEAU_ENV_PREFIX = "CASTOR_TABLEAU_"


class TableauCredentials(BaseSettings):
    """
    Tableau's credentials to connect to both APIs (REST and GRAPHQL)
    """

    model_config = SettingsConfigDict(
        env_prefix=TABLEAU_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    server_url: str
    site_id: str = ""

    password: Optional[str] = None
    token: Optional[str] = None
    token_name: Optional[str] = None
    user: Optional[str] = None

    @field_validator("site_id", mode="before")
    @classmethod
    def _check_site_id(cls, site_id: Optional[str]) -> str:
        if not site_id or site_id.lower() == _DEFAULT_SITE_ID_USER_INPUT:
            return _DEFAULT_SERVER_SITE_ID
        return site_id

    @model_validator(mode="after")
    def _check_user_xor_pat_login(self) -> "TableauCredentials":
        """
        Checks that credentials are correctly input, it means either:
        - User and password are filled
        - Token and Token name are filled
        """
        user_login = self.password and self.user
        pat_login = self.token_name and self.token
        if not user_login and not pat_login:
            raise ValueError("Either token or user identification is required")
        if user_login and pat_login:
            raise ValueError("Can't have both token and user identification")
        return self



from .auth import Auth, BasicAuth, BearerAuth, CustomAuth
from .client import APIClient
from .pagination import FetchNextPageBy, PaginationModel, fetch_all_pages
from .safe_request import RequestSafeMode, ResponseJson, handle_response
from .utils import build_url



import logging
import threading
import time
from typing import Optional

import requests

from ....utils import (
    BearerAuth,
    build_url,
    handle_response,
)
from .endpoints import (
    SigmaEndpointFactory,
)

logger = logging.getLogger(__name__)


_AUTH_TIMEOUT_S = 60
_REFRESH_BUFFER_S = 300


class SigmaBearerAuth(BearerAuth):
    def __init__(self, host: str, token_payload: dict[str, str]):
        auth_endpoint = SigmaEndpointFactory.authentication()
        self.authentication_url = build_url(host, auth_endpoint)
        self.token_payload = token_payload
        self._token_expires_at: Optional[float] = None
        self._token_lock = threading.Lock()

    def fetch_token(self) -> str:
        """Returns the token and sets its expiration time."""
        token_api_path = self.authentication_url
        token_response = requests.post(
            token_api_path, data=self.token_payload, timeout=_AUTH_TIMEOUT_S
        )
        response_data = handle_response(token_response)
        expires_in_seconds = int(response_data["expires_in"])
        self._token_expires_at = time.time() + expires_in_seconds
        return response_data["access_token"]

    def _is_token_expired_or_expiring_soon(self) -> bool:
        """
        Returns True if the token is expired or will expire soon (within buffer time)
        """
        if self._token_expires_at is None:
            return False

        return time.time() >= (self._token_expires_at - _REFRESH_BUFFER_S)

    def _needs_refresh(self, force_refresh: bool = False) -> bool:
        """Returns True if the token needs to be refreshed."""
        is_expired = self._is_token_expired_or_expiring_soon()
        return not self._token or force_refresh or is_expired

    def _fetch_token(self, force_refresh: bool = False) -> Optional[str]:
        """Returns the API token, refreshing it if needed (thread-safe)."""
        if not self._needs_refresh(force_refresh):
            return f"Bearer {self._token}"

        with self._token_lock:
            if not self._needs_refresh(force_refresh):
                return f"Bearer {self._token}"

            logger.info("Refreshing authentication token...")
            self._token = self.fetch_token()
            return f"Bearer {self._token}"



import logging
from collections.abc import Iterable
from typing import Optional, Union

from looker_sdk.sdk.api40.models import LookmlModel

from ...logger import add_logging_file_handler, set_stream_handler_to_stdout
from ...utils import (
    SafeMode,
    StreamableList,
    current_timestamp,
    deep_serialize,
    get_output_filename,
    write_json,
    write_summary,
)
from .api import (
    ApiClient,
    ExtractionParameters,
    LookerCredentials,
    lookml_explore_names,
)
from .assets import LookerAsset
from .multithreading import MultithreadingFetcher

logger = logging.getLogger(__name__)


def _extract_explores_by_name(
    lookmls: Iterable[LookmlModel], client: ApiClient
) -> Iterable[dict]:
    explore_names = lookml_explore_names(lookmls)
    explores = client.explores(explore_names)
    for explore in explores:
        yield deep_serialize(explore)  # type: ignore


def _safe_mode(
    extraction_parameters: ExtractionParameters,
) -> Optional[SafeMode]:
    if extraction_parameters.is_safe_mode:
        return None
    add_logging_file_handler(extraction_parameters.output)
    return SafeMode((Exception,), float("inf"))


def _client(
    credentials: LookerCredentials,
    safe_mode: Optional[SafeMode],
    page_size: int,
) -> ApiClient:
    return ApiClient(
        credentials=credentials, safe_mode=safe_mode, page_size=page_size
    )


def iterate_all_data(
    client: ApiClient,
    search_per_folder: bool,
    thread_pool_size: int,
    log_to_stdout: bool,
) -> Iterable[Union[StreamableList, tuple[LookerAsset, list]]]:
    """Iterate over the extracted Data From looker"""

    logger.info("Extracting users from Looker API")
    users = client.users()
    yield LookerAsset.USERS, deep_serialize(users)

    logger.info("Extracting folders from Looker API")
    folders = client.folders()
    folder_ids: set[str] = {folder.id for folder in folders if folder.id}
    yield LookerAsset.FOLDERS, deep_serialize(folders)

    logger.info("Extracting looks from Looker API")
    fetcher = MultithreadingFetcher(
        folder_ids=folder_ids,
        client=client,
        thread_pool_size=thread_pool_size,
        log_to_stdout=log_to_stdout,
    )
    if search_per_folder:
        looks_stream = fetcher.fetch_assets(LookerAsset.LOOKS)
        yield LookerAsset.LOOKS, StreamableList(looks_stream)
    else:
        yield LookerAsset.LOOKS, deep_serialize(client.looks())

    logger.info("Extracting dashboards from Looker API")
    if search_per_folder:
        dashboards_stream = fetcher.fetch_assets(LookerAsset.DASHBOARDS)
        yield LookerAsset.DASHBOARDS, StreamableList(dashboards_stream)
    else:
        dashboards = client.dashboards()
        yield LookerAsset.DASHBOARDS, deep_serialize(dashboards)

    logger.info("Extracting lookml models from Looker API")
    lookmls = client.lookml_models()
    yield LookerAsset.LOOKML_MODELS, deep_serialize(lookmls)

    logger.info("Extracting explores from Looker API")
    explores = _extract_explores_by_name(lookmls, client)
    yield LookerAsset.EXPLORES, StreamableList(explores)
    del lookmls

    logger.info("Extracting connections from Looker API")
    yield LookerAsset.CONNECTIONS, deep_serialize(client.connections())

    logger.info("Extracting projects from Looker API")
    yield LookerAsset.PROJECTS, deep_serialize(client.projects())

    logger.info("Extracting groups hierarchy from Looker API")
    groups_hierarchy = client.groups_hierarchy()
    yield LookerAsset.GROUPS_HIERARCHY, deep_serialize(groups_hierarchy)

    logger.info("Extracting groups roles from Looker API")
    yield LookerAsset.GROUPS_ROLES, deep_serialize(client.groups_roles())

    logger.info("Extracting content views from Looker API")
    yield LookerAsset.CONTENT_VIEWS, deep_serialize(client.content_views())

    logger.info("Extracting users attributes from Looker API")
    users_attributes = client.users_attributes()
    yield LookerAsset.USERS_ATTRIBUTES, deep_serialize(users_attributes)


def extract_all(**kwargs) -> None:
    """
    Extract Data From looker and store it locally in files under the
    output_directory
    """
    extraction_parameters = ExtractionParameters(**kwargs)
    output_directory = extraction_parameters.output

    credentials = LookerCredentials(**kwargs)

    if extraction_parameters.log_to_stdout:
        set_stream_handler_to_stdout()

    safe_mode = _safe_mode(extraction_parameters)
    client = _client(
        credentials=credentials,
        safe_mode=safe_mode,
        page_size=extraction_parameters.page_size,
    )

    ts = current_timestamp()

    data = iterate_all_data(
        client=client,
        search_per_folder=extraction_parameters.search_per_folder,
        thread_pool_size=extraction_parameters.thread_pool_size,
        log_to_stdout=extraction_parameters.log_to_stdout,
    )
    for asset, data in data:
        filename = get_output_filename(asset.value, output_directory, ts)
        write_json(filename, data)

    write_summary(output_directory, ts, base_url=credentials.base_url)
